# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from collections import defaultdict
import threading
import time

from oslo_config import cfg
from oslo_log import log
import oslo_messaging
from vitrage.utils.datetime import format_utcnow

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import GraphAction
from vitrage.common.constants import VertexProperties as VProps

from vitrage.datasources.consistency import CONSISTENCY_DATASOURCE
from vitrage.datasources import utils
from vitrage import messaging

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class DriverExec(object):

    def __init__(self, process_output_func, persist, graph):
        self.graph = graph
        self.process_output_func = process_output_func
        self.persist = persist

    def snapshot_get_all(self, action=DatasourceAction.INIT_SNAPSHOT):
        driver_names = CONF.datasources.types
        LOG.info('get_all starting for %s', driver_names)
        t1 = time.time()
        events_count = 0
        for d in driver_names:
            events_count += self.get_all(d, action)
        LOG.info('get_all and processing took %s for %s events',
                 time.time() - t1, events_count)
        self.persist.store_graph()

    def get_all(self, driver_name, action):
        try:
            LOCK_BY_DRIVER.acquire(driver_name)
            driver = utils.get_drivers_by_name([driver_name])[0]
            LOG.info("run driver get_all: %s", driver_name)
            now = format_utcnow()
            events = driver.get_all(action)
            count = self.process_output_func(events)
            LOG.info("run driver get_all: %s done (%s events)",
                     driver_name, count)

            if driver.should_delete_outdated_entities():
                vertices_to_delete = \
                    self._find_vertices_to_delete(driver_name, now)
                if vertices_to_delete:
                    delete_events = self._to_events(vertices_to_delete)
                    count_deleted = self.process_output_func(delete_events)
                    LOG.info("run driver delete outdated vertices: %s done "
                             "(%s events)", driver_name, count_deleted)
                    count += count_deleted
            return count
        except Exception:
            LOG.exception("run driver get_all: %s Failed", driver_name)
        finally:
            LOCK_BY_DRIVER.release(driver_name)
        return 0

    def get_changes(self, driver_name):
        if not LOCK_BY_DRIVER.acquire(driver_name, blocking=False):
            LOG.info("%s get_changes canceled during get_all execution",
                     driver_name)
            return 0
        try:
            driver = utils.get_drivers_by_name([driver_name])[0]
            LOG.info("run driver get_changes: %s", driver_name)
            events = driver.get_changes(DatasourceAction.UPDATE)
            count = self.process_output_func(events)
            LOG.info("run driver get_changes: %s done (%s events)",
                     driver_name, count)
            return count
        except Exception:
            LOG.exception("run driver get_changes: %s Failed", driver_name)
        finally:
            LOCK_BY_DRIVER.release(driver_name)
        return 0

    @staticmethod
    def _to_events(vertices):
        return (
            {
                DSProps.ENTITY_TYPE: CONSISTENCY_DATASOURCE,
                DSProps.DATASOURCE_ACTION: DatasourceAction.UPDATE,
                DSProps.SAMPLE_DATE: format_utcnow(),
                DSProps.EVENT_TYPE: GraphAction.DELETE_ENTITY,
                VProps.VITRAGE_ID: vertex[VProps.VITRAGE_ID],
                VProps.ID: vertex.get(VProps.ID, None),
                VProps.VITRAGE_TYPE: vertex[VProps.VITRAGE_TYPE],
                VProps.VITRAGE_CATEGORY: vertex[VProps.VITRAGE_CATEGORY],
                VProps.IS_REAL_VITRAGE_ID: True
            }
            for vertex in vertices
        )

    def _find_vertices_to_delete(self, driver_name, now):
        query = {
            'and': [
                {'==': {VProps.VITRAGE_DATASOURCE_NAME: driver_name}},
                {'<': {VProps.VITRAGE_SAMPLE_TIMESTAMP: now}},
                {'==': {VProps.VITRAGE_IS_DELETED: False}},
            ]
        }

        return self.graph.get_vertices(query_dict=query)


class DriversNotificationEndpoint(object):

    def __init__(self, processor_func):
        self._processor_func = processor_func
        self._enrich_event_methods = defaultdict(list)

    def init(self):
        driver_names = utils.get_push_drivers_names()
        push_drivers = utils.get_drivers_by_name(driver_names)
        for driver in push_drivers:
            for event in driver.get_event_types():
                self._enrich_event_methods[event].append(driver.enrich_event)
        return self

    def get_listener(self):
        topics = CONF.datasources.notification_topics
        exchange = CONF.datasources.notification_exchange
        transport = messaging.get_transport()
        targets = [oslo_messaging.Target(exchange=exchange, topic=topic)
                   for topic in topics]

        return messaging.get_notification_listener(transport, targets, [self])

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        enrich_event_methods = self._enrich_event_methods[event_type]
        events = []
        for enrich_event_method in enrich_event_methods:
            result = enrich_event_method(payload, event_type)
            if isinstance(result, list):
                events += result
            else:
                events.append(result)
        events = [x for x in events if x is not None]
        LOG.info('EVENTS ENQUEUED:[%s] [%s] \n%s', publisher_id,
                 event_type, events)
        self._processor_func(events)


class LockByDriver(object):

    def __init__(self):
        self.lock_by_driver = dict()

    def acquire(self, driver_name, blocking=True):
        if not self.lock_by_driver.get(driver_name):
            self.lock_by_driver[driver_name] = threading.Lock()
        return self.lock_by_driver[driver_name].acquire(blocking)

    def release(self, driver_name):
        self.lock_by_driver[driver_name].release()


LOCK_BY_DRIVER = LockByDriver()



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from vitrage.datasources.cinder.volume import CINDER_VOLUME_DATASOURCE
from vitrage.datasources.neutron.network import NEUTRON_NETWORK_DATASOURCE
from vitrage.datasources.neutron.port import NEUTRON_PORT_DATASOURCE
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources.nova.zone import NOVA_ZONE_DATASOURCE

OPENSTACK_CLUSTER = 'openstack.cluster'

# Register options for the service
OPTS = [
    cfg.ListOpt('types',
                default=[NOVA_HOST_DATASOURCE,
                         NOVA_INSTANCE_DATASOURCE,
                         NOVA_ZONE_DATASOURCE,
                         CINDER_VOLUME_DATASOURCE,
                         NEUTRON_PORT_DATASOURCE,
                         NEUTRON_NETWORK_DATASOURCE,
                         ],
                help='Names of supported data sources'),
    cfg.ListOpt('path',
                default=['vitrage.datasources'],
                help='base path for data sources'),
    cfg.IntOpt('snapshots_interval',
               default=600,
               min=10,
               help='Time to wait between subsequent datasource snapshots'),
    cfg.IntOpt('snapshot_interval_on_fault',
               default=20,
               min=1,
               help='Time to wait until retrying to snapshot the datasource'
                    ' in case of fault'),
    cfg.ListOpt('notification_topics',
                default=['vitrage_notifications'],
                help='Vitrage configured notifications topic',
                deprecated_name='notification_topic'),
    cfg.StrOpt('notification_exchange',
               required=False,
               help='Exchange that is used for notifications.'),
]



# Copyright 2020 - Inspur - Qitao
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,  software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND,  either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.datasources.cetus.cetus_driver_base import CetusDriverBase
from vitrage.datasources.cetus.cluster import CETUS_CLUSTER_DATASOURCE


class ClusterDriver(CetusDriverBase):

    def get_all(self, datasource_action):
        return self.make_pickleable(self._prepare_entities(
            self.list_all()),
            CETUS_CLUSTER_DATASOURCE,
            datasource_action)

    @staticmethod
    def _prepare_entities(pods):
        clusters_dict = {}
        for pod in pods:

            cluster = {
                "name": pod['cluster']["name"],
                "id": pod['cluster']["cluster_id"],
                "status": pod['cluster']["status"],
            }

            node_id = pod["metadata"]["labels"]["instance_id"]
            cluster_id = pod['cluster']["cluster_id"]
            if cluster_id not in clusters_dict:
                clusters_dict[cluster_id] = cluster
                clusters_dict[cluster_id]['nodes'] = []
            if node_id not in clusters_dict[cluster_id]['nodes']:
                clusters_dict[cluster_id]['nodes'].append(node_id)
        return [c for c in clusters_dict.values()]



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from vitrage.common.exception import VitrageError
from vitrage.evaluator.condition import get_condition_common_targets
from vitrage.evaluator.condition import parse_condition
from vitrage.evaluator.condition import SymbolResolver
from vitrage.evaluator.template_data import EdgeDescription
from vitrage.evaluator.template_data import ENTITY
from vitrage.evaluator.template_data import RELATIONSHIP
from vitrage.evaluator.template_data import Scenario
from vitrage.evaluator.template_fields import TemplateFields as TFields
from vitrage.evaluator.template_loading.subgraph_builder import \
    SubGraphBuilder
from vitrage.graph import Vertex

LOG = log.getLogger(__name__)


class ScenarioLoader(object):

    def __init__(self, template_schema, name, entities, relationships):
        self.name = name
        self._template_schema = template_schema
        self._template_entities = entities
        self._template_relationships = relationships

        self.entities = {}
        self.relationships = {}
        self.valid_target = None

    def build_scenarios(self, scenarios_defs):
        scenarios = []
        for counter, scenario_def in enumerate(scenarios_defs):
            scenario_id = "%s-scenario%s" % (self.name, counter)
            scenario_dict = scenario_def[TFields.SCENARIO]
            condition = parse_condition(scenario_dict[TFields.CONDITION])
            self.valid_target = calculate_action_target(
                condition,
                self._template_entities,
                self._template_relationships)
            actions = self._build_actions(scenario_dict[TFields.ACTIONS],
                                          scenario_id)
            subgraphs = SubGraphBuilder.from_condition(
                condition,
                self._extract_var_and_update_index)

            scenarios.append(
                Scenario(scenario_id, self._template_schema.version(),
                         condition, actions, subgraphs,
                         self.entities, self.relationships))

        return scenarios

    @classmethod
    def build_equivalent_scenario(cls, scenario, template_id, entity_props):
        entities = scenario.entities.copy()
        entities[template_id] = Vertex(
            vertex_id=entities[template_id].vertex_id,
            properties={k: v for k, v in entity_props})
        relationships = {
            rel_id: _build_equivalent_relationship(rel,
                                                   template_id,
                                                   entity_props)
            for rel_id, rel in scenario.relationships.items()}

        def extract_var(symbol_name):
            if symbol_name in entities:
                return entities[symbol_name], ENTITY
            elif symbol_name in relationships:
                return relationships[symbol_name], RELATIONSHIP
            else:
                raise VitrageError('invalid symbol name: {}'
                                   .format(symbol_name))

        subgraphs = SubGraphBuilder.from_condition(
            scenario.condition, extract_var)

        return Scenario(id=scenario.id + '_equivalence',
                        version=scenario.version,
                        condition=scenario.condition,
                        actions=scenario.actions,
                        subgraphs=subgraphs,
                        entities=entities,
                        relationships=relationships)

    def _build_actions(self, actions_def, scenario_id):
        actions = []

        for counter, action_def in enumerate(actions_def):
            action_id = '%s-action%s' % (scenario_id, counter)
            action_type = action_def[TFields.ACTION][TFields.ACTION_TYPE]
            action_loader = self._template_schema.loaders.get(action_type)

            if action_loader:
                actions.append(action_loader.load(action_id, self.valid_target,
                                                  action_def))
            else:
                LOG.warning('Failed to load action of type %s', action_type)

        return actions

    def _extract_var_and_update_index(self, symbol_name):
        if symbol_name in self._template_relationships:
            relationship = self._template_relationships[symbol_name]
            self.relationships[symbol_name] = relationship
            self.entities.update({
                relationship.edge.source_id: relationship.source,
                relationship.edge.target_id: relationship.target
            })
            return relationship, RELATIONSHIP

        entity = self._template_entities[symbol_name]
        self.entities[symbol_name] = entity
        return entity, ENTITY


def calculate_action_target(condition, entities, relationships):
    """Return a vertex that can be used as an action target.

    External actions like execute_mistral do not have an explicit
    action target. This parameter is a must for the sub-graph matching
    algorithm. If it is missing, we would like to select an arbitrary
    target from the condition.
    If the target result is empty the condition is not valid.

    """
    definition_index = entities.copy()
    definition_index.update(relationships)
    targets = get_condition_common_targets(
        condition,
        definition_index,
        TemplateDataSymbolResolver())
    return {TFields.TARGET: targets.pop()} if targets else None


class TemplateDataSymbolResolver(SymbolResolver):
    def is_relationship(self, symbol):
        return isinstance(symbol, EdgeDescription)

    def get_relationship_source_id(self, relationship):
        return relationship.source.vertex_id

    def get_relationship_target_id(self, relationship):
        return relationship.target.vertex_id

    def get_entity_id(self, entity):
        return entity.vertex_id


def _build_equivalent_relationship(relationship,
                                   template_id,
                                   entity_props):
    source = relationship.source
    target = relationship.target
    if relationship.edge.source_id == template_id:
        source = Vertex(vertex_id=source.vertex_id,
                        properties={k: v for k, v in entity_props})
    elif relationship.edge.target_id == template_id:
        target = Vertex(vertex_id=target.vertex_id,
                        properties={k: v for k, v in entity_props})
    return EdgeDescription(source=source,
                           target=target,
                           edge=relationship.edge)



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,  software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND,  either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.datasources.alarm_driver_base import AlarmDriverBase
from vitrage.datasources.ceilometer import CEILOMETER_DATASOURCE
from vitrage.datasources.ceilometer.properties \
    import CeilometerEventType as CeilEventType
from vitrage.datasources.ceilometer.properties \
    import CeilometerProperties as CeilProps
from vitrage.datasources.ceilometer.properties \
    import CeilometerState as CeilState
from vitrage import os_clients
from vitrage.utils import datetime as datetime_utils

LOG = log.getLogger(__name__)


class CeilometerDriver(AlarmDriverBase):

    def __init__(self):
        super(CeilometerDriver, self).__init__()
        self._client = None
        self._init_aodh_event_actions()
        self._cache_all_alarms()

    @property
    def client(self):
        if not self._client:
            self._client = os_clients.ceilometer_client()
        return self._client

    def _vitrage_type(self):
        return CEILOMETER_DATASOURCE

    def _alarm_key(self, alarm):
        return alarm[CeilProps.ALARM_ID]

    def _cache_all_alarms(self):
        alarms = self._get_alarms()
        self._filter_and_cache_alarms(alarms,
                                      self._filter_get_valid)

    def _get_alarms(self):
        try:
            aodh_alarms = self.client.alarms.list()
            return [self._convert_alarm(alarm) for alarm in
                    aodh_alarms if alarm is not None]
        except Exception:
            LOG.exception("Failed to get all alarms.")
        return []

    def _is_erroneous(self, alarm):
        return alarm and alarm[CeilProps.STATE] == CeilState.ALARM

    def _status_changed(self, alarm1, alarm2):
        return alarm1 and alarm2 and \
            not alarm1[CeilProps.STATE] == alarm2[CeilProps.STATE]

    def _is_valid(self, alarm):
        return True

    @classmethod
    def _convert_event_alarm(cls, alarm):
        res = cls._convert_base_alarm(alarm)
        res[CeilProps.EVENT_TYPE] = alarm.event_rule[CeilProps.EVENT_TYPE],
        res[CeilProps.RESOURCE_ID] = _parse_query(alarm.event_rule,
                                                  CeilProps.EVENT_RESOURCE_ID)
        return res

    @classmethod
    def _convert_threshold_alarm(cls, alarm):
        res = cls._convert_base_alarm(alarm)
        res[CeilProps.STATE_TIMESTAMP] = alarm.state_timestamp
        res[CeilProps.RESOURCE_ID] = _parse_query(alarm.threshold_rule,
                                                  CeilProps.RESOURCE_ID)
        return res

    @classmethod
    def _convert_gnocchi_resources_threshold(cls, alarm):
        res = cls._convert_base_alarm_gnocchi(alarm)
        if type(alarm) is not dict:
            alarm = alarm.to_dict()
            res[CeilProps.STATE_TIMESTAMP] = \
                alarm.get(CeilProps.STATE_TIMESTAMP)
            res[CeilProps.RESOURCE_ID] = \
                alarm.get(CeilProps.GNOCCHI_RESOURCES_THRESHOLD_RULE,
                          {}).get(CeilProps.RESOURCE_ID)
        else:
            res[CeilProps.STATE_TIMESTAMP] = \
                alarm.get(CeilProps.DETAIL, {}).get(CeilProps.STATE_TIMESTAMP)
            res[CeilProps.RESOURCE_ID] = \
                alarm.get(CeilProps.DETAIL,
                          {}).get(CeilProps.RULE,
                                  {}).get(CeilProps.RESOURCE_ID)
        return res

    @classmethod
    def _convert_vitrage_alarm(cls, alarm):
        res = cls._convert_base_alarm(alarm)
        res[CeilProps.VITRAGE_ID] = _parse_query(alarm.event_rule,
                                                 CeilProps.VITRAGE_ID)
        res[CeilProps.RESOURCE_ID] = _parse_query(alarm.event_rule,
                                                  CeilProps.RESOURCE_ID)
        return res

    @staticmethod
    def _convert_base_dict_alarm_gnocchi(alarm):
        detail = alarm.get(CeilProps.DETAIL)
        return {
            CeilProps.SEVERITY: alarm.get(CeilProps.SEVERITY),
            CeilProps.PROJECT_ID: alarm.get(CeilProps.PROJECT_ID),
            CeilProps.TIMESTAMP: alarm.get(CeilProps.TIMESTAMP),
            CeilProps.TYPE: alarm.get(CeilProps.TYPE),
            CeilProps.ALARM_ID: alarm.get(CeilProps.ALARM_ID),
            CeilProps.DESCRIPTION: detail.get(CeilProps.DESCRIPTION),
            CeilProps.ENABLED: detail.get(CeilProps.ENABLED),
            CeilProps.NAME: detail.get(CeilProps.NAME),
            CeilProps.REPEAT_ACTIONS: detail.get(CeilProps.REPEAT_ACTIONS),
            CeilProps.STATE: detail.get(CeilProps.STATE)
        }

    @staticmethod
    def _convert_base_non_dict_alarm_gnocchi(alarm):
        alarm = alarm.to_dict()
        return {
            CeilProps.SEVERITY: alarm.get(CeilProps.SEVERITY),
            CeilProps.DESCRIPTION: alarm.get(CeilProps.DESCRIPTION),
            CeilProps.ENABLED: alarm.get(CeilProps.ENABLED),
            CeilProps.ALARM_ID: alarm.get(CeilProps.ALARM_ID),
            CeilProps.NAME: alarm.get(CeilProps.NAME),
            CeilProps.PROJECT_ID: alarm.get(CeilProps.PROJECT_ID),
            CeilProps.REPEAT_ACTIONS: alarm.get(CeilProps.REPEAT_ACTIONS),
            CeilProps.STATE: alarm.get(CeilProps.STATE),
            CeilProps.TIMESTAMP: alarm.get(CeilProps.TIMESTAMP),
            CeilProps.TYPE: alarm.get(CeilProps.TYPE)
        }

    @classmethod
    def _convert_base_alarm_gnocchi(cls, alarm):
        """distinguish between alarm received by notification (type dict)

        to alarm received by _get_alarms() (type alarm).
        """

        if type(alarm) is dict:
            return cls._convert_base_dict_alarm_gnocchi(alarm)

        return cls._convert_base_non_dict_alarm_gnocchi(alarm)

    @staticmethod
    def _convert_base_alarm(alarm):
        return {
            CeilProps.SEVERITY: alarm.severity,
            CeilProps.DESCRIPTION: alarm.description,
            CeilProps.ENABLED: alarm.enabled,
            CeilProps.ALARM_ID: alarm.alarm_id,
            CeilProps.NAME: alarm.name,
            CeilProps.PROJECT_ID: alarm.project_id,
            CeilProps.REPEAT_ACTIONS: alarm.repeat_actions,
            CeilProps.STATE: alarm.state,
            CeilProps.TIMESTAMP: alarm.timestamp,
            CeilProps.TYPE: alarm.type
        }

    @classmethod
    def _convert_alarm(cls, alarm):
        alarm_type = alarm.type
        if alarm_type == CeilProps.EVENT and \
                _is_vitrage_alarm(alarm.event_rule):
            return cls._convert_vitrage_alarm(alarm)
        elif alarm_type == CeilProps.EVENT:
            return cls._convert_event_alarm(alarm)
        elif alarm_type == CeilProps.THRESHOLD:
            return cls._convert_threshold_alarm(alarm)
        elif alarm_type == CeilProps.GNOCCHI_RESOURCES_THRESHOLD:
            return cls._convert_gnocchi_resources_threshold(alarm)
        else:
            LOG.warning('Unsupported Ceilometer alarm type %s' % alarm_type)

    @staticmethod
    def get_event_types():
        # Add event_types to receive notifications about
        return [CeilEventType.CREATION,
                CeilEventType.STATE_TRANSITION,
                CeilEventType.RULE_CHANGE,
                CeilEventType.DELETION]

    def enrich_event(self, event, event_type):
        if event_type in self.actions:
            entity = self.actions[event_type](event)
        else:
            LOG.warning('Unsupported Ceilometer event type %s' % event_type)
            return None

        # Don't need to update entity, only update the cache
        if entity is None:
            return None

        entity[DSProps.EVENT_TYPE] = event_type

        return CeilometerDriver.make_pickleable(
            [entity], CEILOMETER_DATASOURCE,
            DatasourceAction.UPDATE)[0]

    def _init_aodh_event_actions(self):
        self.actions = {
            CeilEventType.CREATION:
                self._convert_alarm_creation_event,
            CeilEventType.RULE_CHANGE:
                self._convert_alarm_rule_change_event,
            CeilEventType.STATE_TRANSITION:
                self._convert_alarm_state_transition_event,
            CeilEventType.DELETION:
                self._convert_alarm_deletion_event
        }

    @classmethod
    def _convert_base_event(cls, event):
        return {
            CeilProps.PROJECT_ID: event[CeilProps.PROJECT_ID],
            CeilProps.ALARM_ID: event[CeilProps.ALARM_ID],
            CeilProps.SEVERITY: event[CeilProps.SEVERITY],
            CeilProps.TIMESTAMP: event[CeilProps.TIMESTAMP],
            CeilProps.USER_ID: event[CeilProps.USER_ID]
        }

    @classmethod
    def _convert_vitrage_alarm_rule(cls, rule):
        return {
            CeilProps.VITRAGE_ID: _parse_query(rule, CeilProps.VITRAGE_ID),
            CeilProps.RESOURCE_ID: _parse_query(rule, CeilProps.RESOURCE_ID)
        }

    @classmethod
    def _convert_threshold_alarm_rule(cls, rule):
        return {
            CeilProps.RESOURCE_ID: _parse_query(rule, CeilProps.RESOURCE_ID),
        }

    @classmethod
    def _convert_gnocchi_resources_threshold_alarm_rule(cls, rule):
        return {
            CeilProps.RESOURCE_ID: _parse_query(rule, CeilProps.RESOURCE_ID),
        }

    @classmethod
    def _convert_event_alarm_rule(cls, rule):
        return {
            CeilProps.EVENT_TYPE: rule[CeilProps.EVENT_TYPE],
            CeilProps.RESOURCE_ID:
                _parse_query(rule, CeilProps.EVENT_RESOURCE_ID)
        }

    @classmethod
    def _convert_detail_event(cls, event):
        alarm_info = event[CeilProps.DETAIL]
        alarm_rule = alarm_info[CeilProps.RULE]

        entity_detail = {
            CeilProps.DESCRIPTION: alarm_info[CeilProps.DESCRIPTION],
            CeilProps.ENABLED: alarm_info[CeilProps.ENABLED],
            CeilProps.NAME: alarm_info[CeilProps.NAME],
            CeilProps.STATE: alarm_info[CeilProps.STATE],
            CeilProps.REPEAT_ACTIONS: alarm_info[CeilProps.REPEAT_ACTIONS],
            CeilProps.TYPE: alarm_info[CeilProps.TYPE],
            CeilProps.STATE_TIMESTAMP: alarm_info[CeilProps.STATE_TIMESTAMP],
            CeilProps.STATE_REASON: alarm_info[CeilProps.STATE_REASON]
        }

        if _is_vitrage_alarm(alarm_rule):
            entity_detail.update(cls._convert_vitrage_alarm_rule(alarm_rule))
        elif entity_detail[CeilProps.TYPE] == CeilProps.EVENT:
            entity_detail.update(cls._convert_event_alarm_rule(alarm_rule))
        elif entity_detail[CeilProps.TYPE] == CeilProps.THRESHOLD:
            entity_detail.update(
                cls._convert_threshold_alarm_rule(alarm_rule))
        elif entity_detail[CeilProps.TYPE] == \
                CeilProps.GNOCCHI_RESOURCES_THRESHOLD:
            entity_detail.update(
                cls._convert_gnocchi_resources_threshold_alarm_rule(
                    alarm_rule))

        return entity_detail

    @classmethod
    def _parse_changed_rule(cls, change_rule):
        entity = {}
        if CeilProps.EVENT_TYPE in change_rule:
            entity[CeilProps.EVENT_TYPE] = change_rule[CeilProps.EVENT_TYPE]
        if 'query' in change_rule:
            event_resource_id = \
                _parse_query(change_rule, CeilProps.EVENT_RESOURCE_ID)
            resource_id = \
                _parse_query(change_rule, CeilProps.RESOURCE_ID)
            if event_resource_id or resource_id:
                entity[CeilProps.RESOURCE_ID] = event_resource_id if \
                    event_resource_id is not None else resource_id

        return entity

    @staticmethod
    def should_delete_outdated_entities():
        return True

    def _convert_alarm_creation_event(self, event):
        entity = self._convert_base_event(event)
        detail = self._convert_detail_event(event)
        entity.update(detail)

        return self._filter_and_cache_alarm(entity, None,
                                            self._filter_get_erroneous,
                                            datetime_utils.utcnow(False))

    def _convert_alarm_rule_change_event(self, event):
        """handle alarm rule change notification

        example of changed rule:
        "detail": {"severity": "critical",
                   "rule":
                       {"query": [{"field": "traits.resource_id",
                                   "type": "",
                                   "value": "1",
                                   "op": "eq"}],
                       "event_type": "instance.update"}}
        """
        old_alarm = self._old_alarm(event)
        entity = old_alarm.copy()

        changed_rule = event[CeilProps.DETAIL]
        for (changed_type, changed_info) in changed_rule.items():
            # handle changed rule which may effect the neighbor
            if changed_type == CeilProps.RULE:
                entity.update(self._parse_changed_rule(
                    changed_rule[changed_type]))
            # handle other changed alarm properties
            elif changed_type in vars(CeilProps).values():
                entity[changed_type] = changed_info

        return self._filter_and_cache_alarm(entity, old_alarm,
                                            self._filter_get_erroneous,
                                            datetime_utils.utcnow(False))

    def _convert_alarm_state_transition_event(self, event):

        old_alarm = self._old_alarm(event)
        entity = old_alarm.copy()
        try:
            entity[CeilProps.STATE] = event[CeilProps.DETAIL][CeilProps.STATE]
        except Exception:
            LOG.exception("Failed to Convert alarm state transition event.")

        return self._filter_and_cache_alarm(entity, old_alarm,
                                            self._filter_get_change,
                                            datetime_utils.utcnow(False))

    def _convert_alarm_deletion_event(self, event):
        alarm_key = self._alarm_key(event)
        alarm = self.cache.pop(alarm_key)[0]
        return alarm if self._is_erroneous(alarm) else None


def _parse_query(data, key):
    query_fields = data.get(CeilProps.QUERY, {})
    for query in query_fields:
        field = query['field']
        if field == key:
            return query['value']
    return None


def _is_vitrage_alarm(rule):
    return _parse_query(rule, CeilProps.VITRAGE_ID) is not None



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log
from osprofiler import profiler

from vitrage.api_handler.apis import base
from vitrage.common.constants import HistoryProps as HProps
from vitrage.common.constants import TenantProps
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.alarm_properties import AlarmProperties as AProps
from vitrage.graph.driver.networkx_graph import NXGraph
from vitrage.graph import Edge
from vitrage.graph import Vertex
from vitrage.storage import db_time

LOG = log.getLogger(__name__)


@profiler.trace_cls("rca apis",
                    info={}, hide_args=False, trace_private=False)
class RcaApis(base.EntityGraphApisBase):

    @base.lock_graph
    def get_rca(self, ctx, root, all_tenants):
        LOG.debug("RcaApis get_rca - root: %s, all_tenants=%s",
                  root, all_tenants)

        project_id = ctx.get(TenantProps.PROJECT_ID, None)
        is_admin_project = ctx.get(TenantProps.IS_ADMIN, False)

        if all_tenants:
            db_nodes, db_edges = self.db.history_facade.alarm_rca(root)
        else:
            db_nodes, db_edges = self.db.history_facade.alarm_rca(
                root,
                project_id=project_id,
                admin=is_admin_project)

        for n in db_nodes:
            start_timestamp = \
                self.db.history_facade.add_utc_timezone(n.start_timestamp)
            n.payload[HProps.START_TIMESTAMP] = str(start_timestamp)
            if n.end_timestamp <= db_time():
                end_timestamp = \
                    self.db.history_facade.add_utc_timezone(n.end_timestamp)
                n.payload[HProps.END_TIMESTAMP] = str(end_timestamp)
                # TODO(annarez): implement state change in processor and DB
                n.payload[VProps.STATE] = AProps.INACTIVE_STATE

        vertices = [Vertex(vertex_id=n.vitrage_id, properties=n.payload) for n
                    in db_nodes]
        edges = [Edge(source_id=e.source_id, target_id=e.target_id,
                      label=e.label, properties=e.payload) for e in db_edges]
        rca_graph = NXGraph(vertices=vertices, edges=edges)

        json_graph = rca_graph.json_output_graph(
            inspected_index=self._find_rca_index(rca_graph, root))

        return json_graph

    @staticmethod
    def _find_rca_index(found_graph, root):
        for root_index, vertex in enumerate(found_graph._g):
            if vertex == root:
                return root_index
        return 0



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from collections import namedtuple

from vitrage.evaluator.template_fields import TemplateFields
from vitrage.evaluator.template_schema_factory import TemplateSchemaFactory

Template = namedtuple('Template', ['uuid', 'data', 'date'])
TEMPLATE_LOADER = 'template_loader'
SYNTAX = 'syntax'
CONTENT = 'content'


def get_template_schema(template):
    metadata = template.get(TemplateFields.METADATA)
    if metadata is None:
        return None
    version = metadata.get(TemplateFields.VERSION,
                           TemplateSchemaFactory.DEFAULT_VERSION)
    return TemplateSchemaFactory().template_schema(version)



# Copyright 2016 - Nokia, ZTE
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.resource_transformer_base import \
    ResourceTransformerBase
from vitrage.datasources.static import STATIC_DATASOURCE
from vitrage.datasources.static import StaticFields
from vitrage.datasources import transformer_base
import vitrage.graph.utils as graph_utils

LOG = logging.getLogger(__name__)


class StaticTransformer(ResourceTransformerBase):

    def _create_snapshot_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_update_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_snapshot_neighbors(self, entity_event):
        return self._create_static_neighbors(entity_event)

    def _create_update_neighbors(self, entity_event):
        return self._create_static_neighbors(entity_event)

    def _create_entity_key(self, entity_event):
        entity_id = entity_event[VProps.ID]
        entity_type = entity_event[StaticFields.TYPE]
        key_fields = self._key_values(entity_type, entity_id)
        return transformer_base.build_key(key_fields)

    @staticmethod
    def get_vitrage_type():
        return STATIC_DATASOURCE

    def _create_vertex(self, entity_event):

        entity_type = entity_event[StaticFields.TYPE]
        entity_category = entity_event.get(StaticFields.CATEGORY,
                                           EntityCategory.RESOURCE)
        entity_id = entity_event[VProps.ID]
        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]

        if entity_type in self.transformers:
            properties = {
                VProps.VITRAGE_TYPE: entity_type,
                VProps.ID: entity_id,
                VProps.VITRAGE_CATEGORY: entity_category,
                VProps.VITRAGE_SAMPLE_TIMESTAMP: vitrage_sample_timestamp,
                VProps.VITRAGE_IS_PLACEHOLDER: False
            }
            return self.create_neighbor_placeholder_vertex(**properties)
        else:
            entity_key = self._create_entity_key(entity_event)
            state = entity_event[VProps.STATE]
            update_timestamp = self._format_update_timestamp(
                update_timestamp=None,
                sample_timestamp=vitrage_sample_timestamp)
            metadata = entity_event.get(StaticFields.METADATA, {})

            return graph_utils.create_vertex(
                entity_key,
                vitrage_category=entity_category,
                vitrage_type=entity_type,
                vitrage_sample_timestamp=vitrage_sample_timestamp,
                entity_id=entity_id,
                update_timestamp=update_timestamp,
                entity_state=state,
                metadata=metadata)

    def _create_static_neighbor(self, entity_event, rel):
        if entity_event[StaticFields.STATIC_ID] == rel[StaticFields.SOURCE]:
            neighbor = rel[StaticFields.TARGET]
            is_entity_source = True
        elif entity_event[StaticFields.STATIC_ID] == rel[StaticFields.TARGET]:
            neighbor = rel[StaticFields.SOURCE]
            is_entity_source = False
        else:
            LOG.error("Invalid relationship {} in entity_event {}".format(
                rel, entity_event))
            return None
        if rel[StaticFields.SOURCE] == rel[StaticFields.TARGET]:
            neighbor = entity_event
        return self._create_neighbor(
            entity_event,
            neighbor[VProps.ID],
            neighbor[StaticFields.TYPE],
            rel[StaticFields.RELATIONSHIP_TYPE],
            is_entity_source=is_entity_source)

    def _create_static_neighbors(self, entity_event):
        relationships = entity_event.get(StaticFields.RELATIONSHIPS,
                                         [])
        return [self._create_static_neighbor(entity_event, rel)
                for rel in relationships]



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import GraphAction
from vitrage.datasources.driver_base import DriverBase
from vitrage.utils import datetime as datetime_utils

LOG = log.getLogger(__name__)


class AlarmDriverBase(DriverBase):
    def __init__(self):
        super(DriverBase, self).__init__()
        self.cache = dict()

    def _vitrage_type(self):
        """Return the vitrage_type of the datasource """
        pass

    def _alarm_key(self, alarm):
        """Return a unique key of the alarm, to identify it in the cache """
        pass

    def _get_alarms(self):
        """Return the list of alarms of this plugin """
        pass

    def _enrich_alarms(self, alarms):
        """Optionally add more data to the alarms

        :param alarms: list of alarms to be enriched
        :return:
        """
        pass

    def _is_erroneous(self, alarm):
        """Check if the state of the alarm is erroneous

        :param alarm:
        :return: True/False based on the alarm state
        """
        pass

    def _status_changed(self, new_alarm, old_alarm):
        """Check if the status of the two alarms is different

        :param alarm1:
        :param alarm2:
        :return: True/False based on the alarms states
        """
        pass

    def _is_valid(self, alarm):
        """Check if the alarm is valid

        :param alarm: an alarm to check
        :return: True/False
        """
        pass

    def get_all(self, datasource_action):
        return self.make_pickleable(self._get_all_alarms(),
                                    self._vitrage_type(),
                                    datasource_action)

    def get_changes(self, datasource_action):
        return self.make_pickleable(self._get_changed_alarms(),
                                    self._vitrage_type(),
                                    datasource_action)

    def _get_all_alarms(self):
        alarms = self._get_alarms()
        self._enrich_alarms(alarms)
        return self._filter_and_cache_alarms(
            alarms,
            self._filter_get_erroneous)

    def _get_changed_alarms(self):
        alarms = self._get_alarms()
        self._enrich_alarms(alarms)
        return self._filter_and_cache_alarms(
            alarms,
            self._filter_get_change)

    def _filter_and_cache_alarms(self, alarms, filter_):
        alarms_to_update = []
        now = datetime_utils.utcnow(False)

        for alarm in alarms:
            alarm_key = self._alarm_key(alarm)
            old_alarm = self.cache.get(alarm_key, (None, None))[0]
            if self._filter_and_cache_alarm(alarm, old_alarm, filter_, now):
                alarms_to_update.append(alarm)

        # add alarms that were deleted
        # (i.e. the alarm definition was deleted from the datasource)
        values = list(self.cache.values())
        for cached_alarm, timestamp in values:
            if self._is_erroneous(cached_alarm) and timestamp is not now:
                LOG.debug('deleting cached_alarm %s', cached_alarm)
                cached_alarm[DSProps.EVENT_TYPE] = GraphAction.DELETE_ENTITY
                alarms_to_update.append(cached_alarm)
                self.cache.pop(self._alarm_key(cached_alarm))

        return alarms_to_update

    def _filter_get_valid(self, alarm, _):
        return alarm if self._is_valid(alarm) else None

    def _filter_get_erroneous(self, alarm, old_alarm):
        return alarm \
            if self._is_valid(alarm) and \
            (self._is_erroneous(alarm) or self._is_erroneous(old_alarm)) \
            else None

    def _filter_get_change(self, alarm, old_alarm):
        if not self._is_valid(alarm):
            return None
        if self._status_changed(alarm, old_alarm):
            return alarm
        elif not old_alarm and self._is_erroneous(alarm):
            return alarm
        else:
            return None

    def _filter_and_cache_alarm(self, alarm, old_alarm, filter_, time):
        ret = alarm if filter_(alarm, old_alarm) else None
        self.cache[self._alarm_key(alarm)] = alarm, time
        return ret

    def _old_alarm(self, event):
        alarm_key = self._alarm_key(event)
        return self.cache.get(alarm_key, (None, None))[0]



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import codecs
from collections import namedtuple
import json
from os.path import dirname
from os import walk


def store_specs(spec, target_filename, target_folder=None):
    """Stores specs in JSON format.

    :param spec: specification of trace generator
    :param target_filename: name of file to save in
    :param target_folder: folder for file. defaults to "resources"
    :return: full path to target file
    """

    target = _get_full_path(target_filename, target_folder)
    with open(target, 'w') as outfile:
        json.dump(spec, outfile, sort_keys=True, indent=4, ensure_ascii=False)
    return target


def load_specs(target_filename, target_folder=None):
    """Loads JSON from file in target location. Defaults to "resources" folder

    :param target_filename:
    :param target_folder:
    :return:
    """

    target = _get_full_path(target_filename, target_folder)
    reader = codecs.getreader("utf-8")
    with open(target, "rb") as infile:
        return json.load(reader(infile))


def get_def_templates_dict_from_list(def_temps_list):
    """Turns a list of def_temps into a dictionary of def_temps where the keys

    are their index in the list. Used by unit tests

    :param def_temps_list: def_temp list to convert
    :type def_temps_list: list
    :return: a def_temps dict
    :rtype: dict
    """

    Template = namedtuple('Template', 'data')
    dict = {}
    for num, item in zip(range(len(def_temps_list)), def_temps_list):
        dict[num] = Template(item)

    return dict


def _get_full_path(target_filename, target_folder):
    """Returns the full path for the given folder and filename

    :param target_filename: filename to search. Can be None.
    :param target_folder: folder to search. If None, defaults to "resources"
    :return: full path for the given info
    """

    if target_folder is None:
        target_folder = '%s/mock_configurations/driver' % \
                        get_resources_dir()
    target = '{0}/{1}'.format(target_folder, target_filename)
    return target


def get_resources_dir(filename=None, target_folder='resources'):
    """Locates the resources directory dynamically.

    :param filename: file to locate in directory
    :param target_folder: name of folder.
    :return: path to resources directory
    """

    parent_dir = dirname(dirname(__file__))
    for dir_path, _, file_names in walk(parent_dir):
        if target_folder in dir_path and  \
                (filename is None or filename in file_names):
            return dir_path


def generate_vals(param_specs):
    """Generate values from specs.

    :param param_specs: dictionary where the value is in regex format
    :return: dictionary with generated values for regexs
    """

    if isinstance(param_specs, dict):
        current_info = {k: generate_vals(v) for k, v in param_specs.items()}
    elif isinstance(param_specs, list) or isinstance(param_specs, tuple):
        # convert tuples to lists
        current_info = [generate_vals(param) for param in param_specs]
    elif param_specs:  # assumes primitive type
        current_info = str(param_specs)
    else:
        current_info = None
    return current_info


def merge_vals(current, update):
    """Update the current container with updated values.

    Supports dictionary, list and basic types, as well as nesting.
    :param current: the container to update
    :param update: the container to update with
    :return: updated container
    """

    if current is None:  # value previously not in "current"
        return update
    if isinstance(update, dict):
        for k in update:
            current[k] = merge_vals(current.get(k, None), update[k])
    elif isinstance(update, list):
        # assumes the update is <= from current - use nicely!
        for i in range(len(update)):
            if i < len(current):
                current[i] = merge_vals(current[i], update[i])
            else:
                current.append(update[i])
    else:
        current = update
    return current



# Copyright 2017 - ZTE Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.nagios import NAGIOS_DATASOURCE
from vitrage.datasources.zabbix import ZABBIX_DATASOURCE
from vitrage.evaluator.actions.evaluator_event_transformer \
    import VITRAGE_DATASOURCE
from vitrage.evaluator.template_loading.equivalence_loader import \
    EquivalenceLoader
from vitrage.tests import base
from vitrage.tests.mocks import utils
from vitrage.utils import file as file_utils


class EquivalenceTemplateTest(base.BaseTest):

    BASIC_TEMPLATE = 'basic.yaml'

    def test_equivalence_template(self):

        equivalence_path = '%s/templates/general/equivalences/%s' % (
            utils.get_resources_dir(),
            self.BASIC_TEMPLATE)
        equivalence_definition = file_utils.load_yaml_file(equivalence_path,
                                                           True)
        equivalences = EquivalenceLoader(equivalence_definition).equivalences

        expected = [
            frozenset([
                frozenset([(VProps.VITRAGE_CATEGORY, EntityCategory.ALARM),
                           (VProps.VITRAGE_TYPE, NAGIOS_DATASOURCE),
                           (VProps.NAME, 'host_problem')]),
                frozenset([(VProps.VITRAGE_CATEGORY, EntityCategory.ALARM),
                           (VProps.VITRAGE_TYPE, ZABBIX_DATASOURCE),
                           (VProps.NAME, 'host_fail')]),
                frozenset([(VProps.VITRAGE_CATEGORY, EntityCategory.ALARM),
                           (VProps.VITRAGE_TYPE, VITRAGE_DATASOURCE),
                           (VProps.NAME, 'host_down')])
            ]),
        ]

        self.assertEqual(expected, equivalences)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from oslo_config import cfg
from testtools import matchers

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import GraphAction
from vitrage.datasources.zabbix.properties import ZabbixProperties as ZProps
from vitrage.datasources.zabbix import ZABBIX_DATASOURCE
from vitrage.tests.base import IsEmpty
from vitrage.tests.mocks import utils
from vitrage.tests.unit.datasources.zabbix.mock_driver import MockZabbixDriver
from vitrage.tests.unit.datasources.zabbix.zabbix_base_test import \
    ZabbixBaseTest


# noinspection PyProtectedMember
class ZabbixDriverTest(ZabbixBaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.CONFIG_FILE,
                   help='Zabbix configuration file',
                   default=utils.get_resources_dir()
                        + '/zabbix/zabbix_conf.yaml'),
    ]

    def setUp(self):
        super(ZabbixDriverTest, self).setUp()
        self.conf_reregister_opts(self.OPTS, group=ZABBIX_DATASOURCE)

    def test_get_all(self):
        # Test Setup
        zabbix_driver = MockZabbixDriver()

        alarm_data1 = self._extract_alarm_data(triggerid='1', status='1')
        alarm_data2 = self._extract_alarm_data(triggerid='2', status='1',
                                               value='1')
        alarm_data3 = self._extract_alarm_data(triggerid='3', value='1')
        alarm_data4 = self._extract_alarm_data(triggerid='4')

        zabbix_driver.set_alarm_datas([alarm_data1,
                                       alarm_data2,
                                       alarm_data3,
                                       alarm_data4])
        # Test Action
        alarms = zabbix_driver._get_all_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data3, alarms)

    def test_get_all_functionality(self):

        # Step 1 - Services with status OK should not be returned
        # Test setup scenario
        zabbix_driver = MockZabbixDriver()

        alarm_data1 = self._extract_alarm_data()
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2')
        alarm_data3 = self._extract_alarm_data(z_resource_name='compute-2',
                                               triggerid='2')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        # Test action
        alarms = zabbix_driver._get_all_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, IsEmpty())

        # Step 2 - one raised alarm
        # Test setup
        alarm_data1 = self._extract_alarm_data(value='1')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        # Test action
        alarms = zabbix_driver._get_all_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)

        # Step 3 - two raised alarms
        # Test setup
        alarm_data1 = self._extract_alarm_data(value='1', priority='4')
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               value='1')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        expected_alarm1 = alarm_data1
        expected_alarm2 = copy.copy(alarm_data2)
        expected_alarm2[ZProps.RESOURCE_NAME] = 'host2'

        # Test action
        alarms = zabbix_driver._get_all_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(2))
        self._assert_contains(expected_alarm1, alarms)
        self._assert_contains(expected_alarm2, alarms)

        # Step 4 - Check inactive alarms. Get all function should return
        # inactive alarm (alarm that its status has changed to OK)
        # Test setup
        alarm_data1 = self._extract_alarm_data()
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        expected_alarm1 = alarm_data1
        expected_alarm2 = copy.copy(alarm_data2)
        expected_alarm2[ZProps.RESOURCE_NAME] = 'host2'

        # Test action
        alarms = zabbix_driver._get_all_alarms()

        # Test assertions
        # The alarms of alarm_data1/2 should be returned although their
        # status is OK, because they were not OK earlier
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(2))
        self._assert_contains(expected_alarm1, alarms)
        self._assert_contains(expected_alarm2, alarms)

        # Step 4 - get all when all alarms are inactivated and their status
        # was not changed

        # Test action
        alarms = zabbix_driver._get_all_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'alarms is None')
        self.assertThat(alarms, IsEmpty())

    def test_get_changes_functionality(self):

        # Step 1 - get changes when all alarms are OK
        # Test setup
        zabbix_driver = MockZabbixDriver()

        alarm_data1 = self._extract_alarm_data(priority='2')
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               priority='2')
        alarm_data3 = self._extract_alarm_data(z_resource_name='compute-2',
                                               description='Uptime',
                                               priority='3')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        # Test action
        alarms = zabbix_driver._get_changed_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, IsEmpty())

        # Step 2 - get changes when alarm is raised
        # Test setup
        alarm_data1 = self._extract_alarm_data(priority='2', value='1')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        # Test action
        alarms = zabbix_driver._get_changed_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)

        # Step 3 - get changes when the priority of inactive alarm is changed
        # Test setup
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               priority='3')
        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        # Test action
        alarms = zabbix_driver._get_changed_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, IsEmpty())

        # Step 4 - get changes when:
        # 1. alarm1 - priority of active alarm is changed (should be returned)
        # 2. alarm2 - raised alarm (should be returned)
        # 3. alarm3 - priority of inactive alarm is changed (should not
        #             be returned)
        # Test setup
        alarm_data1 = self._extract_alarm_data(priority='4', value='1')
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               priority='1', value='1')
        alarm_data3 = self._extract_alarm_data(z_resource_name='compute-2',
                                               triggerid='22222',
                                               priority='1')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        expected_alarm1 = alarm_data1
        expected_alarm2 = copy.copy(alarm_data2)
        expected_alarm2[ZProps.RESOURCE_NAME] = 'host2'

        # Test action
        alarms = zabbix_driver._get_changed_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(2))
        self._assert_contains(expected_alarm1, alarms)
        self._assert_contains(expected_alarm2, alarms)

        # Step 5 - get changes when all active alarms are changed to inactive
        # Test setup
        alarm_data1 = self._extract_alarm_data(priority='4')
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               priority='1')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        expected_alarm1 = alarm_data1
        expected_alarm2 = copy.copy(alarm_data2)
        expected_alarm2[ZProps.RESOURCE_NAME] = 'host2'

        # Test action
        alarms = zabbix_driver._get_changed_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(2))
        self._assert_contains(expected_alarm1, alarms)
        self._assert_contains(expected_alarm2, alarms)

        # Step 6 - get changes when no change occurred
        # Action
        alarms = zabbix_driver._get_changed_alarms()

        # Test assertions
        self.assertIsNotNone(alarms, 'alarms is None')
        self.assertThat(alarms, IsEmpty())

    def test_get_changes_and_get_all(self):

        # Step 1 - get changes
        # Step setup
        zabbix_driver = MockZabbixDriver()

        alarm_data1 = self._extract_alarm_data(priority='2', value='1')
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               priority='2')
        alarm_data3 = self._extract_alarm_data(z_resource_name='compute-2',
                                               triggerid='2')
        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        # Step action
        alarms = zabbix_driver._get_changed_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)

        # Step 2 - get changes when no change occurred (returns nothing)
        # Step action
        alarms = zabbix_driver._get_changed_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, IsEmpty())

        # Step 3 - get all
        # Step action
        alarms = zabbix_driver._get_all_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)

        # Step 4 - get all for second time
        # (when no change has occurred it returns the same)
        # Step action
        alarms = zabbix_driver._get_all_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)

        # Step 5 - calling get changes right after get all (returns nothing)
        # Step setup
        alarm_data1 = self._extract_alarm_data(priority='4', value='1')
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               priority='1',
                                               value='1')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        expected_alarm1 = alarm_data1
        expected_alarm2 = copy.copy(alarm_data2)
        expected_alarm2[ZProps.RESOURCE_NAME] = 'host2'

        # Step action
        get_all_alarms = zabbix_driver._get_all_alarms()
        changed_alarms = zabbix_driver._get_changed_alarms()

        # Step assertions
        self.assertIsNotNone(get_all_alarms, 'No alarms returned')
        self.assertThat(get_all_alarms, matchers.HasLength(2))
        self._assert_contains(expected_alarm1, get_all_alarms)
        self._assert_contains(expected_alarm2, get_all_alarms)

        self.assertIsNotNone(changed_alarms, 'No alarms returned')
        self.assertThat(changed_alarms, IsEmpty())

        # Step 6 - get changes
        # Step setup
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2',
                                               priority='4',
                                               value='1')
        alarm_data3 = self._extract_alarm_data(z_resource_name='compute-2',
                                               triggerid='2',
                                               priority='4',
                                               value='1')

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])

        expected_alarm1 = copy.copy(alarm_data2)
        expected_alarm1[ZProps.RESOURCE_NAME] = 'host2'
        expected_alarm2 = copy.copy(expected_alarm1)
        expected_alarm2[ZProps.TRIGGER_ID] = '2'

        # Step action
        alarms = zabbix_driver._get_changed_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(2))
        self._assert_contains(expected_alarm1, alarms)
        self._assert_contains(expected_alarm2, alarms)

    def test_delete_alarm(self):

        # Test setup
        alarm_data1 = self._extract_alarm_data(value='1')
        alarm_data2 = self._extract_alarm_data(z_resource_name='compute-2')
        alarm_data3 = self._extract_alarm_data(z_resource_name='compute-2',
                                               triggerid='2')

        # Step 1 - delete inactive alarm
        # Step setup
        zabbix_driver = MockZabbixDriver()

        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2, alarm_data3])
        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2])

        # Step action
        alarms = zabbix_driver._get_all_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)

        # Step 2 - delete active alarm
        # Step setup
        zabbix_driver.set_alarm_datas([alarm_data2])

        # Step action
        alarms = zabbix_driver._get_all_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)
        self.assertEqual(GraphAction.DELETE_ENTITY,
                         alarms[0][DSProps.EVENT_TYPE])

        # Step 3 - get changes after get all should not return deleted alarm
        # Step action
        alarms = zabbix_driver._get_changed_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'alarms is None')
        self.assertThat(alarms, IsEmpty())

        # Step 4 -
        # Step setup
        zabbix_driver.set_alarm_datas([alarm_data1, alarm_data2])
        zabbix_driver._get_all_alarms()
        zabbix_driver.set_alarm_datas([alarm_data2])

        # Step action
        alarms = zabbix_driver._get_changed_alarms()

        # Step assertions
        self.assertIsNotNone(alarms, 'No alarms returned')
        self.assertThat(alarms, matchers.HasLength(1))
        self._assert_contains(alarm_data1, alarms)
        self.assertEqual(GraphAction.DELETE_ENTITY,
                         alarms[0][DSProps.EVENT_TYPE])

    def _extract_alarm_data(self,
                            z_resource_name='compute-1',
                            description='cpu',
                            status='0',
                            value='0',
                            priority='1',
                            triggerid='0'):

        return {ZProps.ZABBIX_RESOURCE_NAME: z_resource_name,
                ZProps.DESCRIPTION: description,
                ZProps.STATUS: status,
                ZProps.VALUE: value,
                ZProps.PRIORITY: priority,
                ZProps.RESOURCE_NAME: z_resource_name,
                ZProps.TRIGGER_ID: triggerid}



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils as utils

from vitrage import keystone_client

CONF = cfg.CONF
LOG = log.getLogger(__name__)

OPTS = [
    cfg.StrOpt('aodh_version', default='2', help='Aodh version'),
    cfg.StrOpt('ceilometer_version', default='2', help='Ceilometer version'),
    cfg.StrOpt('nova_version', default='2.11', help='Nova version'),
    cfg.StrOpt('cinder_version', default='3', help='Cinder version'),
    cfg.StrOpt('glance_version', default='2', help='Glance version'),
    cfg.StrOpt('heat_version', default='1', help='Heat version'),
    cfg.StrOpt('mistral_version', default='2', help='Mistral version'),
    cfg.StrOpt('gnocchi_version', default='1', help='Gnocchi version'),
    cfg.StrOpt('trove_version', default='1', help='Trove version'),
    cfg.StrOpt('monasca_version', default='2_0', help='Monasca version'),
    cfg.BoolOpt('use_nova_versioned_notifications',
                default=True,
                help='Indicates whether to use Nova versioned notifications.'
                     'The default is True. If False, the deprecated Nova '
                     'legacy notifications will be used.'
                     'This flag must be set to False if notification_format '
                     'is set to "unversioned" in nova.conf'),
]

_client_modules = {
    'aodh': 'aodhclient.client',
    'ceilometer': 'ceilometerclient.client',
    'nova': 'novaclient.client',
    'cinder': 'cinderclient.client',
    'glance': 'glanceclient.client',
    'neutron': 'neutronclient.v2_0.client',
    'heat': 'heatclient.client',
    'mistral': 'mistralclient.api.v2.client',
    'gnocchi': 'gnocchiclient.v1.client',
    'trove': 'troveclient.v1.client',
    'monasca': 'monascaclient.client'
}


def driver_module(driver):
    mod_name = _client_modules[driver]
    module = utils.import_module(mod_name)
    return module


def gnocchi_client():
    """Get an instance of the gnocchi client"""
    try:
        gn_client = driver_module('gnocchi')
        client = gn_client.Client(
            session=keystone_client.get_session())
        LOG.info('Gnocchi client created')
        return client
    except Exception:
        LOG.exception('Create Gnocchi client - Got Exception')


def aodh_client():
    """Get an instance of aodh client"""
    try:
        ao_client = driver_module('aodh')
        client = ao_client.Client(
            CONF.aodh_version,
            session=keystone_client.get_session())
        LOG.info('Aodh client created')
        return client
    except Exception:
        LOG.exception('Create Aodh client - Got Exception.')


def ceilometer_client():
    """Get an instance of ceilometer client"""
    try:
        cm_client = driver_module('ceilometer')
        client = cm_client.get_client(
            version=CONF.ceilometer_version,
            session=keystone_client.get_session(),
        )
        LOG.info('Ceilometer client created')
        return client
    except Exception:
        LOG.exception('Create Ceilometer client - Got Exception.')


def nova_client():
    """Get an instance of nova client"""
    try:
        n_client = driver_module('nova')
        client = n_client.Client(
            version=CONF.nova_version,
            region_name=CONF.service_credentials.region_name,
            session=keystone_client.get_session(),
        )
        LOG.info('Nova client created')
        return client
    except Exception:
        LOG.exception('Create Nova client - Got Exception.')


def trove_client():
    """Get an instance of trove client"""
    try:
        tr_client = driver_module('trove')
        client = tr_client.Client(
            version=CONF.trove_version,
            region_name=CONF.service_credentials.region_name,
            session=keystone_client.get_session(),
        )
        LOG.info('Trove client created')
        return client
    except Exception:
        LOG.exception('Create Trove client - Got Exception.')


def cinder_client():
    """Get an instance of cinder client"""
    try:
        cin_client = driver_module('cinder')
        client = cin_client.Client(
            version=CONF.cinder_version,
            region_name=CONF.service_credentials.region_name,
            session=keystone_client.get_session(),
        )
        LOG.info('Cinder client created')
        return client
    except Exception:
        LOG.exception('Create Cinder client - Got Exception.')


def glance_client():
    """Get an instance of glance client"""
    try:
        glan_client = driver_module('glance')
        client = glan_client.Client(
            version=CONF.glance_version,
            session=keystone_client.get_session(),
        )
        LOG.info('Glance client created')
        return client
    except Exception:
        LOG.exception('Create Glance client - Got Exception')


def neutron_client():
    """Get an instance of neutron client"""
    try:
        ne_client = driver_module('neutron')
        client = ne_client.Client(
            region_name=CONF.service_credentials.region_name,
            session=keystone_client.get_session()
        )
        LOG.info('Neutron client created')
        return client
    except Exception:
        LOG.exception('Create Neutron client - Got Exception.')


def heat_client():
    """Get an instance of heat client"""
    try:
        he_client = driver_module('heat')
        client = he_client.Client(
            version=CONF.heat_version,
            region_name=CONF.service_credentials.region_name,
            session=keystone_client.get_session()
        )
        LOG.info('Heat client created')
        return client
    except Exception:
        LOG.exception('Create Heat client - Got Exception.')


def mistral_client():
    """Get an instance of Mistral client"""
    try:
        mi_client = driver_module('mistral')
        client = mi_client.Client(
            region_name=CONF.service_credentials.region_name,
            session=keystone_client.get_session(),
        )
        LOG.info('Mistral client created')
        return client
    except Exception:
        LOG.exception('Create Mistral client - Got Exception.')


def zaqar_client():
    """Get an instance of Zaqar client"""
    try:
        z_client = driver_module('zaqar')
        client = z_client.Client(
            session=keystone_client.get_session(),
        )
        LOG.info('Zaqar client created')
        return client
    except Exception:
        LOG.exception('Create Zaqar client - Got Exception.')


def monasca_client():
    """Get an instance of Monasca client"""
    try:
        mon_client = driver_module('monasca')

        session = keystone_client.get_session()
        endpoint = session.get_endpoint(service_type='monitoring',
                                        interface='publicURL')
        client = mon_client.Client(
            api_version=CONF.monasca_version,
            session=session,
            endpoint=endpoint
        )
        LOG.info('Monasca client created')
        return client
    except Exception:
        LOG.exception('Create Monasca client - Got Exception.')



# Copyright 2016 - ZTE, Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging
from testtools import matchers

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import UpdateMethod
from vitrage.datasources.ceilometer import CEILOMETER_DATASOURCE
from vitrage.datasources.ceilometer.properties \
    import CeilometerProperties as CeilProps
from vitrage.datasources.ceilometer.transformer import CeilometerTransformer
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.tests.mocks import mock_transformer as mock_sync
from vitrage.tests.unit.datasources.ceilometer.\
    ceilometer_transformer_base_test \
    import CeilometerTransformerBaseTest

LOG = logging.getLogger(__name__)


class TestCeilometerAlarmTransformer(CeilometerTransformerBaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PULL),
    ]

    # noinspection PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(TestCeilometerAlarmTransformer, cls).setUpClass()
        cls.transformers = {}
        cls.conf = cfg.ConfigOpts()
        cls.conf.register_opts(cls.OPTS, group=CEILOMETER_DATASOURCE)
        cls.transformers[CEILOMETER_DATASOURCE] = \
            CeilometerTransformer(cls.transformers)

    def test_key_values_with_vitrage_alarm(self):
        LOG.debug('Ceilometer transformer test: '
                  'get key values(vitrage_alarm)')

        # Test setup
        entity = {CeilProps.VITRAGE_ID: 'test',
                  DSProps.ENTITY_TYPE: CEILOMETER_DATASOURCE,
                  CeilProps.ALARM_ID: '12345'}
        transformer = self.transformers[CEILOMETER_DATASOURCE]

        # Test action
        observed_key_fields = transformer._create_entity_key(entity)

        # Test assertions
        self.assertEqual('test', observed_key_fields)

    def test_key_values(self):
        LOG.debug('Ceilometer transformer test: get key values(aodh alarm)')

        # Test setup
        entity = {DSProps.ENTITY_TYPE: CEILOMETER_DATASOURCE,
                  CeilProps.ALARM_ID: '12345'}
        transformer = self.transformers[CEILOMETER_DATASOURCE]

        # Test action
        entity_key_fields = transformer._create_entity_key(entity).split(":")

        # Test assertions
        self.assertEqual(EntityCategory.ALARM, entity_key_fields[0])
        self.assertEqual(CEILOMETER_DATASOURCE, entity_key_fields[1])
        self.assertEqual(entity[CeilProps.ALARM_ID], entity_key_fields[2])

    def test_snapshot_transform(self):
        LOG.debug('Ceilometer alarm transformer test: '
                  'transform entity event snapshot')

        # Test setup
        spec_list = mock_sync.simple_aodh_alarm_generators(alarm_num=3,
                                                           snapshot_events=3)
        static_events = mock_sync.generate_random_events_list(spec_list)

        for event in static_events:
            # convert neighbor from dict to vertex object
            neighbors = event[TransformerBase.QUERY_RESULT]
            vertices = []
            for neighbor in neighbors:
                neighbor_vertex = self._convert_dist_to_vertex(neighbor)
                vertices.append(self.transformers[CEILOMETER_DATASOURCE].
                                update_uuid_in_vertex(neighbor_vertex))
            event[TransformerBase.QUERY_RESULT] = vertices

            # Test action
            wrapper = self.transformers[CEILOMETER_DATASOURCE].transform(event)

            # Test assertions
            vertex = wrapper.vertex
            self._validate_aodh_vertex_props(vertex, event)

            neighbors = wrapper.neighbors
            self.assertThat(neighbors, matchers.HasLength(1))
            self._validate_neighbors(neighbors, vertex.vertex_id, event)

            self._validate_action(event, wrapper)


class TestCeilometerAlarmPushTransformer(CeilometerTransformerBaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    # noinspection PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(TestCeilometerAlarmPushTransformer, cls).setUpClass()
        cls.transformers = {}
        cls.conf = cfg.ConfigOpts()
        cls.conf.register_opts(cls.OPTS, group=CEILOMETER_DATASOURCE)
        cls.transformers[CEILOMETER_DATASOURCE] = \
            CeilometerTransformer(cls.transformers)

    def test_update_transform(self):
        LOG.debug('Ceilometer update alarm transformer test:'
                  'transform entity event update')

        # Test setup
        spec_list = \
            mock_sync.simple_aodh_update_alarm_generators(alarm_num=5,
                                                          update_events=5)
        static_events = mock_sync.generate_random_events_list(spec_list)

        for event in static_events:
            # convert neighbor from dict to vertex object
            neighbors = event[TransformerBase.QUERY_RESULT]
            vertices = []
            for neighbor in neighbors:
                neighbor_vertex = self._convert_dist_to_vertex(neighbor)
                vertices.append(self.transformers[CEILOMETER_DATASOURCE].
                                update_uuid_in_vertex(neighbor_vertex))
            event[TransformerBase.QUERY_RESULT] = vertices

            # Test action
            wrapper = self.transformers[CEILOMETER_DATASOURCE].transform(event)

            # Test assertions
            vertex = wrapper.vertex
            self._validate_aodh_vertex_props(vertex, event)

            neighbors = wrapper.neighbors
            self.assertThat(neighbors, matchers.HasLength(1))
            self._validate_neighbors(neighbors, vertex.vertex_id, event)

            self._validate_action(event, wrapper)



# Copyright 2018 - Nokia, ZTE
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from vitrage.datasources.static.transformer import StaticTransformer

LOG = logging.getLogger(__name__)


class MockTransformer(StaticTransformer):

    def __init__(self, transformers):
        super(MockTransformer, self).__init__(transformers)

    def _create_vertex(self, entity_event):
        vertex = super(MockTransformer, self)._create_vertex(entity_event)
        for k, v in vertex.items():
            if isinstance(v, list):
                vertex[k] = tuple(v)
        return vertex



# Copyright 2018 - Nokia Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


def db_row_to_dict(row):

    return {
        'id': row.id,
        'project_id': row.project_id,
        'is_admin_webhook': row.is_admin_webhook,
        'created_at': row.created_at,
        'url': row.url,
        'regex_filter': row.regex_filter,
        'headers': row.headers
    }



# Copyright 2020
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.utils import file as file_utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class TmfApi639Config(object):
    def __init__(self):
        try:
            tmfapi639_config_file = CONF.tmfapi639[DSOpts.CONFIG_FILE]
            tmfapi639_config = file_utils.load_yaml_file(tmfapi639_config_file)
            self.endpoints = self._create_mapping(tmfapi639_config)
        except Exception as e:
            LOG.error("Failed initialization: " + str(e))
            self.endpoints = []

    @staticmethod
    def _create_mapping(config):
        """Read URL list from config dictionary"""
        LOG.debug(config)
        endpoint_list = []
        # Tuple list containing either 1 or 2 elements (Endpoint and updates)
        for e in config:
            snapshot_url = e["endpoint"]["snapshot"]
            update_url = ""
            if "update" in e["endpoint"]:
                update_url = e["endpoint"]["update"]
            if update_url != "":
                endpoint_list.append((snapshot_url, update_url))
            else:
                endpoint_list.append(snapshot_url)
        LOG.info("Finished reading endpoints file")
        return endpoint_list



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import UpdateMethod
from vitrage.datasources.nova.instance.driver import InstanceDriver
from vitrage.tests import base
from vitrage.tests.mocks import mock_driver as mock_sync

LOG = logging.getLogger(__name__)


# noinspection PyProtectedMember
class NovaHostTransformerTest(base.BaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    def test_use_versioned_notifications(self):
        LOG.debug('Nova instance driver test')

        # Test setup
        driver = InstanceDriver()
        self.cfg_fixture.config(
            use_nova_versioned_notifications=True)
        update_versioned_event, update_legacy_event = self._create_events()

        # Test action
        events = driver.enrich_event(update_versioned_event,
                                     'instance.create.end')
        self.assert_is_not_empty(events)

        # Test action
        events = driver.enrich_event(update_legacy_event,
                                     'compute.instance.create.end')
        self.assert_is_empty(events)

    def test_use_legacy_notifications(self):
        LOG.debug('Nova instance driver test')

        # Test setup
        driver = InstanceDriver()
        self.cfg_fixture.config(
            use_nova_versioned_notifications=False)
        update_versioned_event, update_legacy_event = self._create_events()

        # Test action
        events = driver.enrich_event(update_versioned_event,
                                     'instance.create.end')
        self.assert_is_empty(events)

        # Test action
        events = driver.enrich_event(update_legacy_event,
                                     'compute.instance.create.end')
        self.assert_is_not_empty(events)

    @staticmethod
    def _create_events():
        spec_list = mock_sync.simple_instance_generators(
            host_num=1, vm_num=1, update_events=1,
            use_nova_versioned_format=True
        )
        update_versioned_event = \
            mock_sync.generate_random_events_list(spec_list)[0]

        spec_list = mock_sync.simple_instance_generators(
            host_num=1, vm_num=1, update_events=1,
            use_nova_versioned_format=False
        )
        update_legacy_event = \
            mock_sync.generate_random_events_list(spec_list)[0]

        return update_versioned_event, update_legacy_event



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from unittest import mock

import oslo_messaging


from vitrage.tests.functional.api.v1 import FunctionalTest


class StatusTest(FunctionalTest):
    def __init__(self, *args, **kwds):
        super(StatusTest, self).__init__(*args, **kwds)
        self.auth = 'noauth'

    def test_get_status_ok(self):
        with mock.patch('pecan.request') as request:
            client = mock.Mock()
            client.call.return_value = True
            request.client.prepare.return_value = client
            resp = self.get_json('/status/')
            self.assert_dict_equal({'reason': 'OK'}, resp)

    def test_get_status_not_ok(self):
        with mock.patch('pecan.request') as request:
            client = mock.Mock()
            client.call.return_value = False
            request.client.prepare.return_value = client
            resp = self.get_json('/status/', expect_errors=True)
            self.assertEqual(503, resp.status_code)
            self.assertIn('vitrage-graph is not ready', resp.text)

    def test_get_status_not_ok_timeout(self):
        with mock.patch('pecan.request') as request:
            client = mock.Mock()
            client.call.side_effect = oslo_messaging.MessagingTimeout()
            request.client.prepare.return_value = client
            resp = self.get_json('/status/', expect_errors=True)
            self.assertEqual(503, resp.status_code)
            self.assertIn('vitrage-graph is not available', resp.text)



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from vitrage.common.exception import VitrageError

LOG = logging.getLogger(__name__)

operators = [
    '<',
    '<=',
    # '=',
    '==',
    '!=',
    '>=',
    '>',
]

logical_operations = [
    'and',
    'or'
]


def create_predicate(query_dict):
    """Create predicate from a logical and/or/==/>/etc expression

    Example Input:
    --------------
    query_dict = {
        'and': [
            {'==': {'CATEGORY': 'ALARM'}},
            {'or': [
                {'>': {'TIME': 150}},
                {'==': {'VITRAGE_IS_DELETED': True}}
            ]}
        ]
    }

    Example Output:
    --------------
    lambda item: ((item['CATEGORY']== 'ALARM') and
                  ((item['TIME']> 150) or (item['VITRAGE_IS_DELETED']== True)))

    Example Usage:
    --------------
    match = create_predicate(query_dict)
    if match(vertex):
        print vertex

    :param query_dict:
    :return: a predicate "match(item)"
    """
    try:
        expression = _create_query_expression(query=query_dict)
        LOG.debug('create_predicate::%s', expression)
        expression = 'lambda item: ' + expression
        return eval(expression)
    except Exception as e:
        LOG.error('invalid query format %s. Exception: %s',
                  query_dict, e)
        raise VitrageError('invalid query format %s. Exception: %s',
                           query_dict, e)


def _create_query_expression(query, parent_operator=None):
    expressions = []

    # First element or element under logical operation
    if not parent_operator and isinstance(query, dict):
        (key, value) = query.copy().popitem()
        return _create_query_expression(value, key)

    # Continue recursion on logical (and/or) operation
    elif parent_operator in logical_operations and isinstance(query, list):
        for val in query:
            expressions.append(_create_query_expression(val))
        return _join_logical_operator(parent_operator, expressions)

    # Recursion evaluate leaf (stop condition)
    elif parent_operator in operators:
        for key, val in query.items():
            expressions.append('item.get(' + _evaluable_str(key) + ')' +
                               parent_operator + ' ' + _evaluable_str(val))
        return _join_logical_operator('and', expressions)
    else:
        raise VitrageError('invalid partial query format',
                           parent_operator, query)


def _evaluable_str(value):
    """wrap string/unicode with back tick"""
    if isinstance(value, str):
        return '\'' + value + '\''
    else:
        return str(value)


def _join_logical_operator(op, expressions):
    """Create an expressions string

    Example input:
        op='AND'
        expressions=['a == b', 'c < d']
    Example output: (a == b AND c < d)
    """
    separator = ' ' + op + ' '
    return '(' + separator.join(expressions) + ')'



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_utils import importutils as utils

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import UpdateMethod
from vitrage.datasources.nagios import NAGIOS_DATASOURCE
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources import utils as ds_utils
from vitrage.datasources.zabbix import ZABBIX_DATASOURCE
from vitrage.tests import base


ZABBIX_DATASOURCE_NONE = '_'.join((ZABBIX_DATASOURCE, UpdateMethod.NONE))
ZABBIX_DATASOURCE_PULL = '_'.join((ZABBIX_DATASOURCE, UpdateMethod.PULL))
ZABBIX_DATASOURCE_PUSH = ZABBIX_DATASOURCE
ZABBIX_DATASOURCE_PULL_NO_INTERVAL = \
    '_'.join((ZABBIX_DATASOURCE, UpdateMethod.PULL, 'no_interval'))


class DatasourceUpdateMethod(base.BaseTest):

    DATASOURCES_OPTS = [
        cfg.ListOpt('types',
                    default=[NOVA_HOST_DATASOURCE,
                             NOVA_INSTANCE_DATASOURCE,
                             NAGIOS_DATASOURCE,
                             ZABBIX_DATASOURCE_NONE,
                             ZABBIX_DATASOURCE_PULL,
                             ZABBIX_DATASOURCE_PUSH,
                             ZABBIX_DATASOURCE_PULL_NO_INTERVAL],
                    help='Names of supported data sources'),
        cfg.StrOpt('notification_topic',
                   default='vitrage_notifications',
                   help='Vitrage configured notifications topic')
    ]

    NOVA_HOST_OPTS = [
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.nova.host.driver.HostDriver',
                   help='Nova host driver class path',
                   required=True),
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.NONE,
                   help='None: updates only via Vitrage periodic snapshots.'
                        'Pull: updates every [changes_interval] seconds.'
                        'Push: updates by getting notifications from the'
                        ' datasource itself.',
                   required=True),
    ]

    NOVA_INSTANCE_OPTS = [
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.nova.instance.driver.'
                           'InstanceDriver',
                   help='Nova instance driver class path',
                   required=True),
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH,
                   required=True),
    ]

    NAGIOS_OPTS = [
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.nagios.driver.NagiosDriver',
                   help='Nagios driver class path',
                   required=True),
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PULL,
                   required=True),
        cfg.IntOpt(DSOpts.CHANGES_INTERVAL,
                   default=30,
                   min=30,
                   help='interval between checking changes in nagios'
                        ' data source'),
    ]

    ZABBIX_OPTS_PUSH = [
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.zabbix.driver.ZabbixDriver',
                   help='Zabbix driver class path',
                   required=True),
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH,
                   required=True),
        cfg.IntOpt(DSOpts.CHANGES_INTERVAL,
                   default=30,
                   min=30,
                   help='interval between checking changes in zabbix'
                        ' data source'),
    ]

    ZABBIX_OPTS_PULL = [
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.zabbix.driver.ZabbixDriver',
                   help='Zabbix driver class path',
                   required=True),
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PULL,
                   required=True),
        cfg.IntOpt(DSOpts.CHANGES_INTERVAL,
                   default=30,
                   min=30,
                   help='interval between checking changes in zabbix'
                        ' data source'),
    ]

    ZABBIX_OPTS_PULL_NO_INTERVAL = [
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.zabbix.driver.ZabbixDriver',
                   help='Zabbix driver class path',
                   required=True),
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PULL,
                   required=True),
    ]

    ZABBIX_OPTS_NONE = [
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.zabbix.driver.ZabbixDriver',
                   help='Zabbix driver class path',
                   required=True),
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.NONE,
                   required=True),
        cfg.IntOpt(DSOpts.CHANGES_INTERVAL,
                   default=30,
                   min=30,
                   help='interval between checking changes in zabbix'
                        ' data source'),
    ]

    def setUp(self):
        super(DatasourceUpdateMethod, self).setUp()
        self.conf_reregister_opts(self.DATASOURCES_OPTS, 'datasources')
        self.conf_reregister_opts(self.NOVA_HOST_OPTS, NOVA_HOST_DATASOURCE)
        self.conf_reregister_opts(self.NOVA_INSTANCE_OPTS,
                                  NOVA_INSTANCE_DATASOURCE)
        self.conf_reregister_opts(self.NAGIOS_OPTS, NAGIOS_DATASOURCE)
        self.conf_reregister_opts(self.ZABBIX_OPTS_NONE,
                                  ZABBIX_DATASOURCE_NONE)
        self.conf_reregister_opts(self.ZABBIX_OPTS_PULL,
                                  ZABBIX_DATASOURCE_PULL)
        self.conf_reregister_opts(self.ZABBIX_OPTS_PUSH,
                                  ZABBIX_DATASOURCE_PUSH)
        self.conf_reregister_opts(self.ZABBIX_OPTS_PULL_NO_INTERVAL,
                                  ZABBIX_DATASOURCE_PULL_NO_INTERVAL)

    def test_datasource_update_method_none(self):
        none_drivers = tuple(driver for driver in self.conf.datasources.types
                             if self.conf[driver].update_method
                             == UpdateMethod.NONE)
        self.assertSequenceEqual(none_drivers,
                                 (NOVA_HOST_DATASOURCE,
                                  ZABBIX_DATASOURCE_NONE))

    def test_datasource_update_method_push(self):
        driver_names = ds_utils.get_push_drivers_names()
        push_drivers = ds_utils.get_drivers_by_name(driver_names)
        self.assertSequenceEqual({utils.import_class(
            self.conf[NOVA_INSTANCE_DATASOURCE].driver), utils.import_class(
            self.conf[ZABBIX_DATASOURCE_PUSH].driver)},
            set(d.__class__ for d in push_drivers))

    def test_datasource_update_method_pull(self):
        driver_names = ds_utils.get_pull_drivers_names()
        self.assertSequenceEqual(
            set([NAGIOS_DATASOURCE, ZABBIX_DATASOURCE_PULL]),
            set(driver_names))

    def test_datasources_notification_topic(self):
        self.assertEqual('vitrage_notifications',
                         self.conf.datasources.notification_topic)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg
from oslo_log import log
import oslo_messaging

from vitrage.common.constants import EdgeLabel as ELabel
from vitrage.common.constants import EdgeProperties as EProps
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import NotifierEventTypes
from vitrage.common.constants import VertexProperties as VProps
from vitrage.evaluator.actions import evaluator_event_transformer as evaluator
from vitrage.graph.driver.networkx_graph import edge_copy
from vitrage.graph.driver.networkx_graph import vertex_copy
from vitrage.messaging import get_transport

CONF = cfg.CONF

LOG = log.getLogger(__name__)


class GraphNotifier(object):
    """Allows writing to message bus"""
    def __init__(self):
        self.oslo_notifier = None
        topics = self._get_topics()
        if not topics:
            LOG.info('Graph Notifier is disabled')
            return
        self.oslo_notifier = oslo_messaging.Notifier(
            get_transport(),
            driver='messagingv2',
            publisher_id='vitrage.graph',
            topics=topics)

    @property
    def enabled(self):
        return self.oslo_notifier is not None

    def _get_topics(self):
        topics = []

        try:
            notifier_topic = CONF.entity_graph.notifier_topic
            notifier_plugins = CONF.notifiers
            if notifier_topic and notifier_plugins:
                topics.append(notifier_topic)
        except Exception:
            LOG.exception('Graph Notifier - missing configuration')

        try:
            machine_learning_topic = \
                CONF.machine_learning.machine_learning_topic
            machine_learning_plugins = CONF.machine_learning.plugins
            if machine_learning_topic and machine_learning_plugins:
                topics.append(machine_learning_topic)
        except Exception:
            LOG.info('Machine Learning - missing configuration')

        return topics

    def notify_when_applicable(self, before, current, is_vertex, graph):
        curr = current
        notification_types = \
            self._get_notification_type(before, curr, is_vertex)
        if not notification_types:
            return

        # in case the vertex point to some resource add the resource to the
        # notification (useful for deduce alarm notifications)
        if curr.get(VProps.VITRAGE_RESOURCE_ID):
            curr = vertex_copy(curr.vertex_id, curr.properties)
            curr.properties[VProps.RESOURCE] = graph.get_vertex(
                curr.get(VProps.VITRAGE_RESOURCE_ID))

        LOG.debug('notification_types : %s', notification_types)
        LOG.debug('notification properties : %s', curr.properties)

        for notification_type in notification_types:
            try:
                self.oslo_notifier.info(
                    {},
                    notification_type,
                    curr.properties)
            except Exception:
                LOG.exception('Cannot notify - %s.', notification_type)

    @staticmethod
    def _get_notification_type(before, current, is_vertex):
        if not is_vertex:
            return None

        notification_types = [
            notification_type(
                before, current, _is_active_deduced_alarm,
                NotifierEventTypes.ACTIVATE_DEDUCED_ALARM_EVENT,
                NotifierEventTypes.DEACTIVATE_DEDUCED_ALARM_EVENT),
            notification_type(
                before, current, _is_active_alarm,
                NotifierEventTypes.ACTIVATE_ALARM_EVENT,
                NotifierEventTypes.DEACTIVATE_ALARM_EVENT),
            notification_type(
                before, current, _is_marked_down,
                NotifierEventTypes.ACTIVATE_MARK_DOWN_EVENT,
                NotifierEventTypes.DEACTIVATE_MARK_DOWN_EVENT),
        ]
        return list(filter(None, notification_types))


class PersistNotifier(object):
    """Allows writing to message bus"""
    def __init__(self):
        self.oslo_notifier = None
        topics = [CONF.persistency.persistor_topic]
        self.oslo_notifier = oslo_messaging.Notifier(
            get_transport(),
            driver='messagingv2',
            publisher_id='vitrage.graph',
            topics=topics)

    def notify_when_applicable(self, before, current, is_vertex, graph):

        curr = current
        notification_types = \
            self._get_notification_type(before, curr, is_vertex)
        if not notification_types:
            return

        # in case the event is on edge, add source and target ids to properties
        # for history
        if not is_vertex:
            curr = edge_copy(
                curr.source_id, curr.target_id, curr.label, curr.properties)
            curr.properties[EProps.SOURCE_ID] = curr.source_id
            curr.properties[EProps.TARGET_ID] = curr.target_id

        LOG.debug('persist_notification_types : %s', notification_types)
        LOG.debug('persist_notification properties : %s', curr.properties)

        for notification_type in notification_types:
            try:
                self.oslo_notifier.info(
                    {},
                    notification_type,
                    curr.properties)
            except Exception:
                LOG.exception('Cannot notify - %s.', notification_type)

    @staticmethod
    def _get_notification_type(before, current, is_vertex):

        notification_types = [
            notification_type(
                before, current, _is_active_alarm,
                NotifierEventTypes.ACTIVATE_ALARM_EVENT,
                NotifierEventTypes.DEACTIVATE_ALARM_EVENT),
            notification_type(
                before, current, _is_active_causes_edge,
                NotifierEventTypes.ACTIVATE_CAUSAL_RELATION,
                NotifierEventTypes.DEACTIVATE_CAUSAL_RELATION),
            NotifierEventTypes.CHANGE_IN_ALARM_EVENT if
            _is_alarm_severity_change(before, current) else None,
            NotifierEventTypes.CHANGE_PROJECT_ID_EVENT if
            _is_resource_project_id_change(before, current) else None,
        ]
        return list(filter(None, notification_types))


def notification_type(before,
                      current,
                      is_active,
                      activate_event_type,
                      deactivate_event_type):
    if not is_active(before):
        if is_active(current):
            return activate_event_type
    else:
        if not is_active(current):
            return deactivate_event_type


def _is_active_deduced_alarm(entity):
    if not entity:
        return False
    if entity.get(VProps.VITRAGE_CATEGORY) == EntityCategory.ALARM and \
            entity.get(VProps.VITRAGE_TYPE) == evaluator.VITRAGE_DATASOURCE:
        return _is_relevant_vertex(entity)
    return False


def _is_active_alarm(entity):
    if entity and entity.get(VProps.VITRAGE_CATEGORY) == EntityCategory.ALARM:
        return _is_relevant_vertex(entity)
    return False


def _is_marked_down(entity):
    if not entity:
        return False
    if entity.get(VProps.VITRAGE_CATEGORY) == EntityCategory.RESOURCE and \
            entity.get(VProps.IS_MARKED_DOWN) is True:
        return _is_relevant_vertex(entity)
    return False


def _is_relevant_vertex(entity):
    if entity.get(VProps.VITRAGE_IS_DELETED, False) or \
            entity.get(VProps.VITRAGE_IS_PLACEHOLDER, False):
        return False
    return True


def _is_active_causes_edge(entity):
    if not entity:
        return False
    if not entity.get(EProps.RELATIONSHIP_TYPE) == ELabel.CAUSES:
        return False
    return not entity.get(EProps.VITRAGE_IS_DELETED)


def _is_alarm_severity_change(before, curr):
    if not (_is_active_alarm(before) and
            _is_active_alarm(curr)):
        return False
    # returns true on activation, deactivation and severity change
    if not before and curr \
            or (before.get(VProps.VITRAGE_OPERATIONAL_SEVERITY) !=
                curr.get(VProps.VITRAGE_OPERATIONAL_SEVERITY)):
        return True
    return False


def _is_resource_project_id_change(before, curr):
    if not (_is_active_alarm(before) and
            _is_active_alarm(curr)):
        return False
    if (before.get(VProps.VITRAGE_RESOURCE_PROJECT_ID) !=
            curr.get(VProps.VITRAGE_RESOURCE_PROJECT_ID)):
        return True
    return False



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import uuid

from oslo_config import cfg

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import UpdateMethod
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.host.transformer import HostTransformer
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources.nova.instance.transformer import InstanceTransformer
from vitrage.datasources.prometheus import PROMETHEUS_DATASOURCE
from vitrage.datasources.prometheus.properties import get_label
from vitrage.datasources.prometheus.properties import PrometheusAlertLabels \
    as PLabels
from vitrage.datasources.prometheus.properties import \
    PrometheusAlertProperties as PProps
from vitrage.datasources.prometheus.properties import PrometheusAlertStatus \
    as PAlertStatus
from vitrage.datasources.prometheus.transformer import PrometheusTransformer
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.tests.mocks import mock_transformer
from vitrage.tests.unit.datasources.test_alarm_transformer_base import \
    BaseAlarmTransformerTest


# noinspection PyProtectedMember
class PrometheusTransformerTest(BaseAlarmTransformerTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    def setUp(self):
        super(PrometheusTransformerTest, self).setUp()
        self.transformers = {}
        self.conf_reregister_opts(self.OPTS, group=PROMETHEUS_DATASOURCE)
        self.transformers[NOVA_HOST_DATASOURCE] = \
            HostTransformer(self.transformers)
        self.transformers[NOVA_INSTANCE_DATASOURCE] = \
            InstanceTransformer(self.transformers)
        self.transformers[PROMETHEUS_DATASOURCE] = \
            PrometheusTransformer(self.transformers)

    def test_create_update_entity_vertex(self):
        # Test setup
        host1 = 'host1'
        instance_id = uuid.uuid4().hex
        event_on_host = self._generate_event_on_host(host1)
        event_on_instance = self._generate_event_on_instance(host1,
                                                             instance_id)
        self.assertIsNotNone(event_on_host)
        self.assertIsNotNone(event_on_instance)

        # Test action
        transformer = self.transformers[PROMETHEUS_DATASOURCE]
        wrapper_for_host = transformer.transform(event_on_host)
        wrapper_for_instance = transformer.transform(event_on_instance)

        # Test assertions
        self._validate_vertex_props(wrapper_for_host.vertex, event_on_host)
        self._validate_vertex_props(wrapper_for_instance.vertex,
                                    event_on_instance)

        # Validate the neighbors: only one valid host neighbor
        host_entity_key = transformer._create_entity_key(event_on_host)
        host_entity_uuid = \
            transformer.uuid_from_deprecated_vitrage_id(host_entity_key)

        instance_entity_key = transformer._create_entity_key(event_on_instance)
        instance_entity_uuid = \
            transformer.uuid_from_deprecated_vitrage_id(instance_entity_key)

        self._validate_host_neighbor(wrapper_for_host,
                                     host_entity_uuid,
                                     host1)

        self._validate_instance_neighbor(wrapper_for_instance,
                                         instance_entity_uuid,
                                         instance_id)

        # Validate the expected action on the graph - update or delete
        self._validate_graph_action(wrapper_for_host)
        self._validate_graph_action(wrapper_for_instance)

    def _validate_vertex_props(self, vertex, event):
        self._validate_alarm_vertex_props(
            vertex, get_label(event, PLabels.ALERT_NAME),
            PROMETHEUS_DATASOURCE, event[DSProps.SAMPLE_DATE])

    def _generate_event_on_host(self, hostname):
        # fake query result to be used by the transformer for determining
        # the neighbor
        query_result = [{VProps.VITRAGE_TYPE: NOVA_HOST_DATASOURCE,
                         VProps.ID: hostname}]
        labels = {PLabels.SEVERITY: 'critical',
                  PLabels.INSTANCE: hostname}

        update_vals = {TransformerBase.QUERY_RESULT: query_result,
                       PProps.LABELS: labels}
        return self._generate_event(update_vals)

    def _generate_event_on_instance(self, hostname, instance_name):
        # fake query result to be used by the transformer for determining
        # the neighbor
        query_result = [{VProps.VITRAGE_TYPE: NOVA_INSTANCE_DATASOURCE,
                         VProps.ID: instance_name}]
        labels = {PLabels.SEVERITY: 'critical',
                  PLabels.INSTANCE: hostname,
                  PLabels.DOMAIN: instance_name}

        update_vals = {TransformerBase.QUERY_RESULT: query_result,
                       PProps.LABELS: labels}
        return self._generate_event(update_vals)

    @staticmethod
    def _generate_event(update_vals):
        generators = mock_transformer.simple_prometheus_alarm_generators(
            update_vals=update_vals)

        return mock_transformer.generate_random_events_list(generators)[0]

    def _is_erroneous(self, vertex):
        return vertex[PProps.STATUS] == PAlertStatus.FIRING



# -*- coding: utf-8 -*-
#
# © 2016 Nokia Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.  You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Collectd plugin for sending notifications to vitrage
"""

import collectd
import hashlib
from vitrage.datasources.collectd import COLLECTD_DATASOURCE

from vitrage.datasources.collectd.collectd_vitrage.plugin import CollectDPlugin
from vitrage.datasources.collectd.collectd_vitrage.plugin import PluginError


from datetime import datetime
from oslo_config import cfg
from oslo_utils import uuidutils

import oslo_messaging as messaging


class VitrageNotifier(CollectDPlugin):
    """Collectd plugin for sending notifications to Vitrage. """

    def configure(self, config, **kwargs):

        super(VitrageNotifier, self).configure(config, **kwargs)

        # to be filled later
        for key in ('transport_url',):
            if key not in self.config.keys():
                message = 'Required configuration key %s missing!' % key
                self.error(message)
                raise PluginError(message)

    def initialize(self):
        """Set up the Vitrage API client and add the notification callback. """

        url = self.config['transport_url']
        transport = messaging.get_notification_transport(cfg.CONF, url)
        self.notifier = messaging.Notifier(transport,
                                           driver='messagingv2',
                                           publisher_id=COLLECTD_DATASOURCE,
                                           topics=['vitrage_notifications'])
        self.add_notification_callback(self.notify)

    def notify(self, notification):
        """Send the notification to Vitrage. """

        # Use a friendly string instead of a number.
        severity = {
            collectd.NOTIF_FAILURE: 'FAILURE',
            collectd.NOTIF_WARNING: 'WARNING',
            collectd.NOTIF_OKAY: 'OK',
        }.get(notification.severity)

        alarm_uuid = self._generate_alarm_id(notification)
        payload = self._create_payload(alarm_uuid, notification, severity)
        notification_id = uuidutils.generate_uuid()

        self.notifier.info(ctxt={'message_id': notification_id,
                                 'publisher_id': COLLECTD_DATASOURCE,
                                 'timestamp': datetime.utcnow()},
                           event_type='collectd.alarm.' + severity.lower(),
                           payload=payload)

        self.info('notification id %r to vitrage: %r' % (notification_id,
                                                         payload))

    @staticmethod
    def _create_payload(alarm_uuid, notification, severity):
        payload = {
            'host': notification.host,
            'plugin': notification.plugin,
            'collectd_type': notification.type,
            'message': notification.message,
            'severity': severity,
            'time': notification.time,
            'id': alarm_uuid
        }
        if notification.plugin_instance:
            payload['plugin_instance'] = notification.plugin_instance
        if notification.type_instance:
            payload['type_instance'] = notification.type_instance
        return payload

    @staticmethod
    def _generate_alarm_id(notification):
        resources = [notification.host, notification.plugin_instance,
                     notification.type_instance, notification.type]
        alarm_id = ''.join([resource for resource in resources if resource])
        alarm_uuid = hashlib.md5(alarm_id.encode("latin-1")).hexdigest()
        return alarm_uuid


# We have to call the constructor in order to actually register our plugin
# with collectd.
VITRAGE = VitrageNotifier()



# Copyright 2019 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import re

from oslo_log import log
from sympy import Not
from sympy import Symbol

from vitrage.evaluator.base import get_template_schema
from vitrage.evaluator import condition as dnf

from vitrage.evaluator.template_fields import TemplateFields
from vitrage.evaluator.template_functions import function_resolver
from vitrage.evaluator.template_functions import GET_PARAM
from vitrage.evaluator.template_functions.v2.functions import get_param
from vitrage.evaluator.template_loading.template_loader_v3 import \
    TemplateLoader as V3TemplateLoader
from vitrage.evaluator.template_validation.base import ValidationError

LOG = log.getLogger(__name__)

RELATION = 'relationship'


class ContentValidator(object):

    @staticmethod
    def validate(template, actual_params):
        _validate_entities_regex(template)
        _validate_conditions(template)
        _validate_parameters(template, actual_params)

        # As part of validation, when it is finished,
        # we try to load the template, as some validations can only be
        # executed at loading phase
        schema = get_template_schema(template)
        V3TemplateLoader().load(schema, template)


def _validate_entities_regex(template):
    for entity in template[TemplateFields.ENTITIES].values():
        for key, value in entity.items():
            if key.lower().endswith(TemplateFields.REGEX):
                try:
                    re.compile(value)
                except Exception:
                    raise ValidationError(47, key, value)


def _validate_conditions(template):
    """Validate conditions

    'alarm_1 [on] host AND host [contains] instance AND alarm_2 [on] instance
     AND host'
    In this condition , replace all occurrences of 'source [label] target'
    so to create :
    'relationship AND relationship AND relationship AND host'
    """
    for scenario in template[TemplateFields.SCENARIOS]:
        condition = scenario[TemplateFields.CONDITION]
        _validate_condition_entity_ids(template, condition)
        _validate_not_condition(condition)


def _validate_parameters(template, actual_params):
    function_resolver.validate_function(
        func_info=function_resolver.FuncInfo(
            name=GET_PARAM, func=get_param, error_code=160),
        template=template,
        actual_params=actual_params)


def _validate_condition_entity_ids(template, condition):
    curr_str = ' ' + condition + ' '

    # Remove all [label] occurrences
    edge_labels_re = r'\s+\[\s*\w+\s*\]\s+'
    curr_str = re.sub(edge_labels_re, ' ', curr_str)
    if '[' in curr_str or ']' in curr_str:
        raise ValidationError(85, condition)

    # Remove all operator occurrences
    operators_re = r'\b(AND|OR|NOT|and|or|not)\b'
    curr_str = re.sub(operators_re, '', curr_str)

    # Remove all entity occurrences
    for entity_id in template[TemplateFields.ENTITIES].keys():
        entity_id_regex = r'\b' + entity_id + r'\b'
        curr_str = re.sub(entity_id_regex, ' ', curr_str)

    # Remove all parentheses occurrences
    curr_str = curr_str.replace('(', '')
    curr_str = curr_str.replace(')', '')

    # Remaining string should be empty
    if curr_str.strip():
        raise ValidationError(10200, condition, curr_str)


def _validate_not_condition(condition):
    """Not operator validation

    1. Not operator can appear only on relationships.
    2. There must be at least one  positive term
    """
    regex = r"(\w+)\s*\[\s*(\w+)\s*\]\s*(\w+)"
    preprocessed_condition = re.sub(regex, RELATION, condition)

    try:
        dnf_condition = dnf.convert_to_dnf_format(preprocessed_condition)
        _validate_not_condition_relationships_recursive(dnf_condition)
        dnf_condition = dnf.parse_condition(preprocessed_condition)
        _validate_positive_term_in_condition(dnf_condition)
    except ValidationError as e:
        raise ValidationError(e.code, e.details, condition)
    except Exception as e:
        raise ValidationError(85, condition, e)


def _validate_not_condition_relationships_recursive(dnf_result):
    if isinstance(dnf_result, Not):
        for arg in dnf_result.args:
            if isinstance(arg, Symbol) and not str(arg).startswith(RELATION):
                raise ValidationError(86, arg)
            else:
                _validate_not_condition_relationships_recursive(arg)
        return

    for arg in dnf_result.args:
        if not isinstance(arg, Symbol):
            _validate_not_condition_relationships_recursive(arg)


def _validate_positive_term_in_condition(dnf_condition):
    if not dnf.is_condition_include_positive_clause(dnf_condition):
        raise ValidationError(134)



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging
from pysnmp.entity.engine import SnmpEngine
from pysnmp.hlapi.asyncore.sync.compat.ntforg import sendNotification
from pysnmp.hlapi.asyncore.transport import UdpTransportTarget
from pysnmp.hlapi.auth import CommunityData
from pysnmp.hlapi.context import ContextData
from pysnmp.proto.rfc1902 import OctetString
from pysnmp.smi.rfc1902 import NotificationType
from pysnmp.smi.rfc1902 import ObjectIdentity
import re

from vitrage.common.constants import VertexProperties as VProps
from vitrage.notifier.plugins.snmp.base import SnmpSenderBase
from vitrage.utils.file import load_yaml_file

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

# TODO(annarez): change NA to N/A
NA = 'NA'
SNMP_TREE = 'snmp_tree'
SEVERITY_MAPPING = 'severity_mapping'
OID = 'oid'
NEXT = 'next'
WITH_VALS = 'with_values'
SEVERITY = 'SEVERITY'
ALARM_OID = 'ALARM_OID'
IP_PAT = re.compile(r'\d+\.\d+\.\d+\.\d+')
PORT_PAT = re.compile(r'\d+')


class SnmpSender(SnmpSenderBase):
    def __init__(self):
        self.hosts = load_yaml_file(CONF.snmp.consumers, True)
        self.oid_tree = load_yaml_file(CONF.snmp.oid_tree, True)
        self.alarm_mapping = \
            load_yaml_file(CONF.snmp.alarm_oid_mapping, True)
        self.oids, self.var_fields = self._build_oids()

    def send_snmp(self, alarm_data):

        alert_details, alert_severity_oid = self._get_details(alarm_data)

        if alert_details:
            alarm_oid = \
                self._get_alert_oid(alert_details[OID], alert_severity_oid)
            if not alarm_oid:
                return
            for host in self.hosts:
                self._send_snmp_trap(host,
                                     self._get_var_binds(alarm_data),
                                     alarm_oid)
        else:
            LOG.info('Vitrage snmp Info: Unrecognized alarm. Alarm type: %s',
                     alarm_data[VProps.NAME])

    def _get_details(self, alarm_data):

        if not (self.hosts and self.alarm_mapping and
                self.oids and self.var_fields):
            LOG.error('Vitrage snmp Error: definitions is '
                      'missing from configuration file')
            return None, None

        alert_severity_oid = self._get_severity_oid(alarm_data)

        if not alert_severity_oid and \
                self.oids.get(SEVERITY):
            LOG.error('Vitrage snmp Error: there '
                      'is no severity mapping in file')
            return None, None

        alarm_name = alarm_data.get(VProps.NAME)
        alert_details = self.alarm_mapping.get(alarm_name)

        return alert_details, alert_severity_oid

    def _build_oids(self):

        if not self.oid_tree:
            return None, None

        oids_dict, var_binds = \
            self._build_oid_recursively('', self.oid_tree[SNMP_TREE],
                                        {}, [], 0)

        oids_dict = {key: oids_dict[key][1:] for key in oids_dict}

        return oids_dict, var_binds

    def _build_oid_recursively(self, oid, curr, oids_dict,
                               var_binds, is_with_val):

        for key in curr:
            new_oid = oid + '.' + curr[key][OID]
            next_node = curr[key].get(NEXT)
            if not next_node:
                if is_with_val:
                    var_binds.append(key)
                oids_dict[key] = new_oid
            else:
                with_val = curr[key].get(WITH_VALS, 0)
                self._build_oid_recursively(new_oid, next_node, oids_dict,
                                            var_binds, with_val)

        return oids_dict, var_binds

    def _get_var_binds(self, alert_values):

        var_binds = [(self.oids[field],
                      OctetString(alert_values.get(field, NA)))
                     for field in self.var_fields]

        return var_binds

    def _get_alert_oid(self, alert_oid, severity_oid):

        sev_oid = self.oids.get(SEVERITY)
        alarm_oid = self.oids.get(ALARM_OID)

        if not (sev_oid or alarm_oid):
            LOG.error("Vitrage snmp Error: snmp tree incorrect format")
            return None

        if severity_oid:
            oid = sev_oid.replace('..', alert_oid + '.' + severity_oid)
            return oid
        else:
            oid = alarm_oid[:-1] + alert_oid

        return oid

    def _get_severity_oid(self, alert_values):

        severity_mapping = self.oid_tree.get(SEVERITY_MAPPING)

        if not severity_mapping:
            return None

        alarm_severity = alert_values.get(VProps.VITRAGE_OPERATIONAL_SEVERITY)
        state = alert_values.get(VProps.STATE)

        if state in severity_mapping:
            return severity_mapping[state]
        elif alarm_severity in severity_mapping:
            return severity_mapping[alarm_severity]

        else:
            LOG.debug('Vitrage snmp Debug: Unsupported alarm severity')
            return None

    @staticmethod
    def _send_snmp_trap(host, var_binds, alarm_oid):

        host_details = host['host']

        send_to = str(host_details.get('send_to'))
        port = str(host_details.get('port', 162))
        community_str = host_details.get('community', 'public')

        if not (send_to and IP_PAT.match(send_to) and PORT_PAT.match(port)):
            LOG.info("Vitrage snmp Info: an SNMP target host was not"
                     " configured correctly")
            return

        LOG.debug("Vitrage snmp Debug: Trap parameters: send_to: %s, "
                  "port: %s, community string: %s" %
                  (send_to, port, community_str))

        error_indication, error_status, error_index, var_bins = next(
            sendNotification(
                SnmpEngine(),
                CommunityData(community_str, mpModel=1),
                UdpTransportTarget((send_to, port)),
                ContextData(),
                'trap',
                NotificationType(
                    ObjectIdentity(alarm_oid),
                ).addVarBinds(*var_binds)
            )
        )

        if error_indication:
            LOG.error('Vitrage snmp Error: Notification not sent: %s' %
                      error_indication)
        elif error_status:
            LOG.error('Vitrage snmp Error: Notification Receiver '
                      'returned error: %s @%s' %
                      (error_status, error_index))



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import timedelta

from concurrent.futures import ThreadPoolExecutor
import dateutil.parser
from futurist import periodics

from oslo_config import cfg
from oslo_log import log
import oslo_messaging as oslo_m
from oslo_utils import timeutils

from vitrage.common.constants import EdgeProperties as EProps
from vitrage.common.constants import ElementProperties as ElementProps
from vitrage.common.constants import HistoryProps as HProps
from vitrage.common.constants import NotifierEventTypes as NETypes
from vitrage.common.constants import VertexProperties as VProps
from vitrage.common.utils import spawn
from vitrage.coordination import service as coord
from vitrage import messaging
from vitrage.storage.sqlalchemy import models
from vitrage.utils.datetime import utcnow

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class PersistorService(coord.Service):
    def __init__(self, worker_id, db_connection):
        super(PersistorService, self).__init__(worker_id)
        self.db_connection = db_connection
        transport = messaging.get_transport()
        target = \
            oslo_m.Target(topic=CONF.persistency.persistor_topic)
        self.listener = messaging.get_notification_listener(
            transport, [target],
            [VitragePersistorEndpoint(self.db_connection)])
        self.scheduler = Scheduler(db_connection)

    def run(self):
        super(PersistorService, self).run()
        LOG.info("Vitrage Persistor Service - Starting...")

        self.listener.start()
        self.scheduler.start_periodic_tasks()

        LOG.info("Vitrage Persistor Service - Started!")

    def terminate(self):
        super(PersistorService, self).terminate()
        LOG.info("Vitrage Persistor Service - Stopping...")

        self.listener.stop()
        self.listener.wait()

        LOG.info("Vitrage Persistor Service - Stopped!")


class VitragePersistorEndpoint(object):
    def __init__(self, db_connection):
        self.db = db_connection
        self.event_type_to_writer = {
            NETypes.ACTIVATE_ALARM_EVENT: self._persist_activated_alarm,
            NETypes.DEACTIVATE_ALARM_EVENT: self._persist_deactivate_alarm,
            NETypes.ACTIVATE_CAUSAL_RELATION: self._persist_activate_edge,
            NETypes.DEACTIVATE_CAUSAL_RELATION: self._persist_deactivate_edge,
            NETypes.CHANGE_IN_ALARM_EVENT: self._persist_change,
            NETypes.CHANGE_PROJECT_ID_EVENT: self._persist_alarm_proj_change,
        }

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        LOG.debug('Event_type: %s Payload %s', event_type, payload)
        self.process_event(event_type, payload)

    def process_event(self, event_type, payload):
        writer = self.event_type_to_writer.get(event_type)
        if not writer:
            LOG.warning('Unrecognized event_type: %s', event_type)
            return
        writer(event_type, payload)

    def _persist_activated_alarm(self, event_type, data):
        event_timestamp = self.event_time(data)

        alarm_row = \
            models.Alarm(
                vitrage_id=data.get(VProps.VITRAGE_ID),
                start_timestamp=event_timestamp,
                name=data.get(VProps.NAME),
                vitrage_type=data.get(VProps.VITRAGE_TYPE),
                vitrage_aggregated_severity=data.get(
                    VProps.VITRAGE_AGGREGATED_SEVERITY),
                vitrage_operational_severity=data.get(
                    VProps.VITRAGE_OPERATIONAL_SEVERITY),
                project_id=data.get(VProps.PROJECT_ID),
                vitrage_resource_type=data.get(VProps.VITRAGE_RESOURCE_TYPE),
                vitrage_resource_id=data.get(VProps.VITRAGE_RESOURCE_ID),
                vitrage_resource_project_id=data.get(
                    VProps.VITRAGE_RESOURCE_PROJECT_ID),
                payload=data)
        self.db.alarms.create(alarm_row)

    def _persist_deactivate_alarm(self, event_type, data):
        vitrage_id = data.get(VProps.VITRAGE_ID)
        event_timestamp = self.event_time(data)
        self.db.alarms.update(
            vitrage_id, HProps.END_TIMESTAMP, event_timestamp)

    def _persist_alarm_proj_change(self, event_type, data):
        vitrage_id = data.get(VProps.VITRAGE_ID)
        self.db.alarms.update(vitrage_id,
                              VProps.VITRAGE_RESOURCE_PROJECT_ID,
                              data.get(VProps.VITRAGE_RESOURCE_PROJECT_ID))

    def _persist_activate_edge(self, event_type, data):
        event_timestamp = self.event_time(data)

        edge_row = \
            models.Edge(
                source_id=data.get(EProps.SOURCE_ID),
                target_id=data.get(EProps.TARGET_ID),
                label=data.get(EProps.RELATIONSHIP_TYPE),
                start_timestamp=event_timestamp,
                payload=data)
        self.db.edges.create(edge_row)

    def _persist_deactivate_edge(self, event_type, data):
        event_timestamp = self.event_time(data)
        source_id = data.get(EProps.SOURCE_ID)
        target_id = data.get(EProps.TARGET_ID)
        self.db.edges.update(
            source_id, target_id, end_timestamp=event_timestamp)

    def _persist_change(self, event_type, data):
        event_timestamp = self.event_time(data)
        change_row = \
            models.Change(
                vitrage_id=data.get(VProps.VITRAGE_ID),
                timestamp=event_timestamp,
                severity=data.get(VProps.VITRAGE_OPERATIONAL_SEVERITY),
                payload=data)
        self.db.changes.create(change_row)

    @staticmethod
    def event_time(data):
        event_timestamp = \
            dateutil.parser.parse(data.get(ElementProps.UPDATE_TIMESTAMP))
        event_timestamp = timeutils.normalize_time(event_timestamp)
        return event_timestamp


class Scheduler(object):

    def __init__(self, db):
        self.db = db
        self.periodic = None

    def start_periodic_tasks(self):
        self.periodic = periodics.PeriodicWorker.create(
            [], executor_factory=lambda: ThreadPoolExecutor(max_workers=10))

        self.add_events_table_expirer_timer()
        self.add_history_tables_expirer_timer()
        spawn(self.periodic.start)

    def add_events_table_expirer_timer(self):
        spacing = 60

        @periodics.periodic(spacing=spacing)
        def expirer_periodic():
            try:
                event_id = self.db.graph_snapshots.query_snapshot_event_id()
                if event_id:
                    LOG.debug('Table events - deleting event id=%s', event_id)
                    self.db.events.delete(event_id)

            except Exception:
                LOG.exception('Table events - periodic cleanup run failed.')

        self.periodic.add(expirer_periodic)
        LOG.info("Table events - periodic cleanup started (%ss)", spacing)

    def add_history_tables_expirer_timer(self):
        spacing = 60

        @periodics.periodic(spacing=spacing)
        def expirer_periodic():
            expire_by = \
                utcnow(with_timezone=False) - \
                timedelta(days=CONF.persistency.alarm_history_ttl)
            try:
                self.db.alarms.delete_expired(expire_by)
            except Exception:
                LOG.exception('History tables - periodic cleanup run failed.')

        self.periodic.add(expirer_periodic)
        LOG.info("History tables - periodic cleanup started (%ss)", spacing)



# Copyright 2016 - ZTE, Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from testtools import matchers

from vitrage.common.constants import DatasourceProperties as DSProp
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.ceilometer import CEILOMETER_DATASOURCE
from vitrage.datasources.ceilometer.properties \
    import CeilometerProperties as CeilProps
from vitrage.datasources import NOVA_HOST_DATASOURCE
from vitrage.datasources import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources import NOVA_ZONE_DATASOURCE
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.tests.functional.datasources.base import \
    TestDataSourcesBase
from vitrage.tests.mocks import mock_transformer


class TestCeilometerAlarms(TestDataSourcesBase):

    DATASOURCES_OPTS = [
        cfg.ListOpt('types',
                    default=[CEILOMETER_DATASOURCE,
                             NOVA_HOST_DATASOURCE,
                             NOVA_INSTANCE_DATASOURCE,
                             NOVA_ZONE_DATASOURCE],
                    help='Names of supported driver data sources'),

        cfg.ListOpt('path',
                    default=['vitrage.datasources'],
                    help='base path for data sources')
    ]

    def setUp(self):
        super(TestCeilometerAlarms, self).setUp()
        self.cfg_fixture.config(group='datasources',
                                types=[
                                    CEILOMETER_DATASOURCE,
                                    NOVA_HOST_DATASOURCE,
                                    NOVA_INSTANCE_DATASOURCE,
                                    NOVA_ZONE_DATASOURCE])
        self.load_datasources()

    def test_ceilometer_alarms_validity(self):
        # Setup
        processor = self._create_processor_with_graph()
        self.assertThat(processor.entity_graph,
                        matchers.HasLength(
                            self._num_total_expected_vertices())
                        )

        detail = {TransformerBase.QUERY_RESULT: '',
                  DSProp.ENTITY_TYPE: CEILOMETER_DATASOURCE}
        spec_list = \
            mock_transformer.simple_aodh_alarm_generators(alarm_num=1,
                                                          snapshot_events=1,
                                                          snap_vals=detail)
        static_events = mock_transformer.generate_random_events_list(spec_list)

        aodh_event = static_events[0]
        aodh_event[CeilProps.RESOURCE_ID] = \
            self._find_entity_id_by_type(processor.entity_graph,
                                         NOVA_HOST_DATASOURCE)

        # Action
        processor.process_event(aodh_event)

        # Test assertions
        self.assertThat(processor.entity_graph,
                        matchers.HasLength(
                            self._num_total_expected_vertices() + 1)
                        )

        aodh_vertices = processor.entity_graph.get_vertices(
            vertex_attr_filter={
                VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
                VProps.VITRAGE_TYPE: CEILOMETER_DATASOURCE
            })
        self.assertThat(aodh_vertices, matchers.HasLength(1))

        aodh_neighbors = processor.entity_graph.neighbors(
            aodh_vertices[0].vertex_id)
        self.assertThat(aodh_neighbors, matchers.HasLength(1))

        self.assertEqual(NOVA_HOST_DATASOURCE,
                         aodh_neighbors[0][VProps.VITRAGE_TYPE])



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.evaluator.template_validation.status_messages import status_msgs
from vitrage.evaluator.template_validation import template_syntax_validator
from vitrage.tests import base
from vitrage.tests.mocks import utils
from vitrage.utils import file as file_utils


class DefTemplateSyntaxValidatorTest(base.BaseTest):

    # noinspection PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(DefTemplateSyntaxValidatorTest, cls).setUpClass()

        cls.def_template_dir_path = utils.get_resources_dir() + \
            '/templates/def_template_tests'

    def test_def_template_with_include_section(self):

        def_template_path = self.def_template_dir_path + \
            '/definition_templates/with_include.yaml'
        def_template = file_utils.load_yaml_file(def_template_path)
        self._test_execution_with_fault_result_for_def_template(def_template,
                                                                143)

    def test_def_template_with_scenario_section(self):

        def_template_path = self.def_template_dir_path + \
            '/definition_templates/with_scenarios.yaml'
        def_template = file_utils.load_yaml_file(def_template_path)
        self._test_execution_with_fault_result_for_def_template(def_template,
                                                                143)

    def test_basic_def_template(self):
        template_path = self.def_template_dir_path +\
            '/templates/basic_with_include.yaml'
        template = file_utils.load_yaml_file(template_path)

        self._test_execution_with_correct_result(template)

    def _test_execution_with_fault_result_for_def_template(self,
                                                           def_template,
                                                           expected_code):

        result = template_syntax_validator.def_template_syntax_validation(
            def_template)

        # Test assertions
        self.assertFalse(result.is_valid_config)
        self.assertTrue(result.comment.startswith(status_msgs[expected_code]))
        self.assertEqual(expected_code, result.status_code)

    def _test_execution_with_correct_result(self, template):

        # Test action
        result = template_syntax_validator.syntax_validation(template)

        # Test assertions
        self.assertTrue(result.is_valid_config)
        self.assertEqual(result.comment, status_msgs[0])
        self.assertEqual(0, result.status_code)



# -*- encoding: utf-8 -*-
# Copyright 2015 - Alcatel-Lucent
# Copyright © 2014-2015 eNovance
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg


def recursive_keypairs(d, separator='.'):
    # taken from ceilometer and gnocchi
    for name, value in sorted(d.items()):
        if isinstance(value, dict):
            for subname, subvalue in recursive_keypairs(value, separator):
                yield ('%s%s%s' % (name, separator, subname), subvalue)
        else:
            yield name, value


def opt_exists(conf_parent, opt):
    try:
        return conf_parent[opt]
    except cfg.NoSuchOptError:
        return False
